"""Helper functions for muP."""

from copy import deepcopy
from pathlib import Path

import pandas as pd
import yaml
from torch import load

from salt.utils.muP_utils.functions_check_muP import (
    get_coord_data,
    get_training_data,
    plot_coord_data,
    plot_training_data,
)

# Paths are hard-coded as the lifetime of these files should last 1 execution
folder_path = "./temp_muP/"
MODEL_MAIN_PATH = "main_model.pt"
MODEL_BASE_PATH = "base_model.pt"
MODEL_DELTA_PATH = "delta_model.pt"
MODEL_BASE_CHECK_PATH = "base_check_model.pt"
MODEL_DELTA_CHECK_PATH = "delta_check_model.pt"
SHAPE_PATH = "storeshapes"
SHAPE_CHECK_PATH = "storeshapes_check"


def get_paths(store_path=None):
    if store_path is None:
        store_path = folder_path
    folder = Path(store_path)
    folder.mkdir(exist_ok=True)
    shape_path = Path(store_path) / SHAPE_PATH
    shape_check_path = Path(store_path) / SHAPE_CHECK_PATH
    model_main_path = Path(store_path) / MODEL_MAIN_PATH
    model_base_path = Path(store_path) / MODEL_BASE_PATH
    model_delta_path = Path(store_path) / MODEL_DELTA_PATH
    model_base_check_path = Path(store_path) / MODEL_BASE_CHECK_PATH
    model_delta_check_path = Path(store_path) / MODEL_DELTA_CHECK_PATH
    return {
        "folder": folder,
        "store_path": store_path,
        "shape_path": shape_path,
        "shape_check_path": shape_check_path,
        "model_main_path": model_main_path,
        "model_base_path": model_base_path,
        "model_delta_path": model_delta_path,
        "model_base_check_path": model_base_check_path,
        "model_delta_check_path": model_delta_check_path,
    }


def recursive_set_val(dictionary, keyTarget, valueSet):
    for key, value in dictionary.items():
        if key == keyTarget:
            return True, key, valueSet
        if isinstance(value, dict):
            found, subkey, subvalue = recursive_set_val(value, keyTarget, valueSet)
            if found:
                value[subkey] = subvalue
                return True, key, value
    return False, None, None


def set_val_nestedKey(dictionary, keyTarget, valueSet):
    """Returns the updated nested dictionary with the (sub-)key keyTarget updated to value valueSet
    Recursively iterates over the keys of nested dictionaries until the keyTarget is found.

    Note: dictionary can be a list of dictionary (useful for init_nets)
    """
    adapt_to_list, found = False, False
    if isinstance(dictionary, list):
        # Some logic to detect if it's a list of dictionary
        adapt_to_list = True
        dictionary = {f"entry_{i}": dictionary[i] for i in range(len(dictionary))}

    # Core step: the recursive search
    for key, value in dictionary.items():
        if key == keyTarget:
            # Key is on the top level of dictionary, no need for recursiveness
            changeKey, changeVal = key, valueSet
            break
        if isinstance(value, dict):
            found, subkey, subvalue = recursive_set_val(value, keyTarget, valueSet)
            if found:
                value[subkey] = subvalue
                changeKey, changeVal = key, value

    if found:
        dictionary[changeKey] = changeVal
        # Further processing if started with list of dictionaries
        if adapt_to_list:
            dictionary = list(dictionary.values())
        return dictionary
    raise KeyError(f"Key {keyTarget} not found in the (nested) dictionary {dictionary}")


def update_config(muP_config, cfg_out, parameter_val):
    for param_config in muP_config.values():
        assert len(param_config["apply_to"]) == len(
            param_config["parameter_name"]
        ), "Define one parameter name per module to apply to."
        param_dic = zip(
            param_config["apply_to"],
            param_config["parameter_name"],
            [param_config[parameter_val]] * len(param_config["apply_to"]),
            strict=True,
        )
        for model_adapt, paramter_adapt, val in param_dic:
            assert (
                model_adapt in cfg_out["model"]["model"]["init_args"]
            ), f"Model key {model_adapt} not found in init_args"
            cfg_out["model"]["model"]["init_args"][model_adapt] = set_val_nestedKey(
                cfg_out["model"]["model"]["init_args"][model_adapt], paramter_adapt, val
            )
    return cfg_out


def generate_base_delta_config(main_config, mod_type):
    """mod_type is either base or delta."""
    with open(main_config) as file:
        cfg = yaml.safe_load(file)
    cfg_out = deepcopy(cfg)

    assert "muP_config" in cfg["model"], "To use the muP, you need to define it in the model config"
    muP_config = cfg["model"]["muP_config"]

    path_dict = get_paths()
    cfg_out_path = (
        Path(path_dict["store_path"]) / "modbase.yaml"
        if mod_type == "base"
        else Path(path_dict["store_path"]) / "./moddelta.yaml"
    )
    parameter_val = "parameter_base" if mod_type == "base" else "parameter_delta"

    shape_path = None
    if "shape_path" in muP_config:
        shape_path = muP_config.pop("shape_path")

    cfg_out = update_config(muP_config, cfg_out, parameter_val)

    # clean out the muP config
    cfg_out["model"]["muP_config"] = {}
    with open(cfg_out_path, "w") as file:
        yaml.dump(cfg_out, file, sort_keys=False)
    return cfg_out_path, shape_path


def generate_config_muPtest(main_config, variations, muP=True):
    """Creates configs with varying dimensions in variations."""
    with open(main_config) as file:
        cfg = yaml.safe_load(file)

    configs = []
    path_dict = get_paths()
    for i, var in enumerate(variations):
        cfg_out = deepcopy(cfg)
        muP_config = {
            "embed_dim": {
                "apply_to": ["init_nets", "encoder"],
                "parameter_name": ["output_size", "embed_dim"],
                "parameter": var,
            }
        }
        cfg_out = update_config(muP_config, cfg_out, "parameter")
        if not muP:
            cfg_out["model"].pop("muP_config")
            remove_muP = {
                "muP": {
                    "apply_to": ["encoder"],
                    "parameter_name": ["muP"],
                    "parameter": False,
                }
            }
            cfg_out = update_config(remove_muP, cfg_out, "parameter")

        cfg_out_path = (
            Path(path_dict["store_path"]) / f"config_muP_{i}.yaml"
            if muP
            else Path(path_dict["store_path"]) / f"config_sP_{i}.yaml"
        )
        configs.append(cfg_out_path)
        with open(cfg_out_path, "w") as file:
            yaml.dump(cfg_out, file, sort_keys=False)
    return configs


def get_model_path(mod_type, path_dict=None):
    if path_dict is None:
        path_dict = get_paths()
    if "temp_" in mod_type:
        return (
            path_dict["model_base_check_path"]
            if "base" in mod_type
            else path_dict["model_delta_check_path"]
            if "delta" in mod_type
            else Path(path_dict["store_path"]) / f"./{mod_type}.pt"
        )
    return (
        path_dict["model_base_path"]
        if "base" in mod_type
        else path_dict["model_delta_path"]
        if "delta" in mod_type
        else path_dict["model_main_path"]
    )


def get_models_muPtest(variations, modInd=None, modType=None):
    models_muP, models_sP = {}, {}
    path_dict = get_paths()
    if modInd is not None:
        if modType is not None:
            if modType:
                models_muP[variations[modInd]] = (
                    Path(path_dict["store_path"]) / f"temp_muP_{modInd}.pt"
                )
            else:
                models_sP[variations[modInd]] = (
                    Path(path_dict["store_path"]) / f"temp_sP_{modInd}.pt"
                )
        else:
            models_muP[variations[modInd]] = Path(path_dict["store_path"]) / f"temp_muP_{modInd}.pt"
            models_sP[variations[modInd]] = Path(path_dict["store_path"]) / f"temp_sP_{modInd}.pt"
        return models_muP, models_sP

    for i, var in enumerate(variations):
        models_muP[var] = Path(path_dict["store_path"]) / f"temp_muP_{i}.pt"
        models_sP[var] = Path(path_dict["store_path"]) / f"temp_sP_{i}.pt"
    return models_muP, models_sP


def load_models(check=False):
    path_dict = get_paths()
    if check:
        return load(path_dict["model_base_check_path"]), load(path_dict["model_delta_check_path"])
    return load(path_dict["model_base_path"]), load(path_dict["model_delta_path"])


def store_shapes_mup(path=None, check=False):
    from mup import make_base_shapes

    path_dict = get_paths(store_path=path)
    store_at = path_dict["shape_path"] if not check else path_dict["shape_check_path"]
    base, delta = load_models(check)
    make_base_shapes(base, delta, store_at)


def instantiate_mup(model, load_from=None, check=False):
    from mup import set_base_shapes

    rescale = not (check)
    path_dict = get_paths(store_path=load_from)
    load_from = str(path_dict["shape_path"]) if not check else str(path_dict["shape_check_path"])
    set_base_shapes(model, load_from, rescale_params=rescale)


def clean_environment():
    path_dict = get_paths()
    path_dict["folder"].unlink()


def coord_check(
    muP,
    models,
    dataloaders,
    optimiser,
    save_path,
    lr=1e-3,
    nsteps=20,
    nseeds=2,
    name_contains=None,
    legend=False,
    mod="GN2",
    save_df=True,
    doPlots=True,
):
    """Coord check for muP. Advised to use a large LR.
    Based on the muP GitHub repo.
    """
    df = get_coord_data(
        models,
        dataloaders,
        mup=muP,
        optimizer=optimiser,
        fix_data=True,
        lr=lr,
        nseeds=nseeds,
        nsteps=nsteps,
        cuda=False,
    )
    if save_df:
        df.to_csv(save_path / "mup.csv" if muP else save_path / "sp.csv")

    if doPlots:
        prm = "μP" if muP else "SP"
        return plot_coord_data(
            df,
            legend=legend,
            name_contains=name_contains,
            save_to=save_path / "mup.png" if muP else save_path / "sp.png",
            suptitle=f"{prm} {mod} with {optimiser} lr={lr} nseeds={nseeds}",
            face_color="xkcd:light grey" if not muP else None,
        )
    return None


def training_check(
    muP,
    models,
    dataloaders,
    optimiser,
    lr,
    nsteps,
    nseeds,
    save_path,
    legend="full",
    mod="GN2",
    save_df=True,
    save_model=None,
    doPlots=True,
    noTQDM=True,
    load_data=None,
):
    """Coord check for muP. Advised to use a large LR.
    Based on the muP GitHub repo.
    """
    if load_data is None:
        df = get_training_data(
            models,
            dataloaders,
            mup=muP,
            optimizer=optimiser,
            lr=lr,
            nseeds=nseeds,
            nsteps=nsteps,
            save_model=save_model,
            cuda=False,
            noTQDM=noTQDM,
        )
        if save_df:
            if save_model is not None:
                width = list(models.keys())[0]
                df.to_csv(save_model / f"width_{width}.csv")
            else:
                df.to_csv(save_path / "mup.csv" if muP else save_path / "sp.csv")
    else:
        df = pd.read_csv(load_data)
    if not doPlots:
        return None

    prm = "μP" if muP else "SP"
    return plot_training_data(
        df,
        legend=legend,
        save_to=save_path / "mup_train.png" if muP else save_path / "sp_train.png",
        suptitle=f"{prm} {mod} with {optimiser} lr={lr} nseeds={nseeds}",
        face_color="xkcd:light grey" if not muP else None,
    )


def check_muP(
    models_muP: dict,
    models_sP: dict,
    lr: float = 1e-2,
    nsteps_training: int = 1500,
    nsteps_coocheck: int = 3,
    nseeds: int = 5,
    batch_size: int = 2000,
    num_workers: int = 10,
    num_train: int = -1,
    train_file: str = "./pp_output_train.h5",
    norm_dict: str = "./norm_dict.yaml",
    save_path: str = "./plots_mu",
    doCoord: bool = True,
    doCoordPlots: bool = True,
    doTraining: bool = True,
    doTrainingPlots: bool = True,
):
    """Script to check muP is well set.

    Parameters
    ----------
    models_muP : dict,
        Dictionary of muP models
        Format: {width: path to model}
    models_sP: dict,
        Dictionary of sP models
        Format: {width: path to model}
    lr : float, default: 1e-2,
        learning rate to use for the tests
    nsteps_training : int, default: 1500,
        nsteps of training check
    nsteps_coocheck : int, default: 3,
        nsteps of coordinate check
    nseeds : int, default: 5,
        number of seeds to use
    batch_size : int, default: 2000,
        mini batch size to use
    num_workers : int, default: 10,
        number of parallel dataloaders to use
    num_train : int, default: -1,
        number of data points to use (-1: all)
    train_file : str, default: "./pp_output_train.h
        path to the training file
    norm_dict : str, default: "./norm_dict.yaml",
        path to the normalisation dictionary
    save_path : str, default: "./plots_mu",
        path to save the results
    doCoord : bool, default: True,
        whether to do the coordinate check
    doCoordPlots : bool, default: True,
        whether to plot the coordinate check results
    doTraining : bool, default: True,
        whether to do the training check
    doTrainingPlots : bool, default: True,
        whether to plot the training check results
    """
    from salt.data.datamodules import SaltDataModule, SaltDataset

    datamodule = SaltDataModule(
        train_file=train_file,
        val_file="",
        batch_size=batch_size,
        num_workers=num_workers,
        num_train=num_train,
        num_val=0,
        num_test=0,
        stage="fit",
    )
    variables = {
        "jets": ["pt_btagJes", "eta_btagJes"],
        "tracks": [
            "d0",
            "z0SinTheta",
            "dphi",
            "deta",
            "qOverP",
            "IP3D_signed_d0_significance",
            "IP3D_signed_z0_significance",
            "phiUncertainty",
            "thetaUncertainty",
            "qOverPUncertainty",
            "numberOfPixelHits",
            "numberOfSCTHits",
            "numberOfInnermostPixelLayerHits",
            "numberOfNextToInnermostPixelLayerHits",
            "numberOfInnermostPixelLayerSharedHits",
            "numberOfInnermostPixelLayerSplitHits",
            "numberOfPixelSharedHits",
            "numberOfPixelSplitHits",
            "numberOfSCTSharedHits",
        ],
    }
    datamodule.train_dset = SaltDataset(
        filename=datamodule.train_file,
        num=datamodule.num_train,
        stage="fit",
        norm_dict=norm_dict,
        variables=variables,
        labels={"jets": ["flavour_label"]},
    )
    dataloader = datamodule.get_dataloader(dataset=datamodule.train_dset, stage="fit", shuffle=True)
    plot_dir = Path(save_path)
    plot_dir.mkdir(exist_ok=True)
    if doCoord:
        coord_check(
            True,
            models_muP,
            dataloader,
            "adamw",
            plot_dir,
            lr=lr,
            nsteps=nsteps_coocheck,
            nseeds=nseeds,
            name_contains=["init_nets", "encoder"],
            doPlots=doCoordPlots,
        )
        coord_check(
            False,
            models_sP,
            dataloader,
            "adamw",
            plot_dir,
            lr=lr,
            nsteps=nsteps_coocheck,
            nseeds=nseeds,
            name_contains=["init_nets", "encoder"],
            doPlots=doCoordPlots,
        )
    if doTraining and models_muP:
        save_muP_models = plot_dir / "trained_muP_models"
        training_check(
            True,
            models_muP,
            dataloader,
            "adamw",
            lr=lr,
            nsteps=nsteps_training,
            nseeds=nseeds,
            save_path=plot_dir,
            save_model=save_muP_models,
            doPlots=doTrainingPlots,
        )
    if doTraining and models_sP:
        save_sP_models = plot_dir / "trained_sP_models"
        training_check(
            False,
            models_sP,
            dataloader,
            "adamw",
            lr=lr,
            nsteps=nsteps_training,
            nseeds=nseeds,
            save_path=plot_dir,
            save_model=save_sP_models,
            doPlots=doTrainingPlots,
        )



import argparse
import shutil
from pathlib import Path


def delete_dirs_without_subdir(folder_path, subdir):
    folder_path = Path(folder_path)

    count = 0
    for directory in folder_path.iterdir():
        if directory.is_dir() and not (directory / subdir).is_dir():
            print(f"Deleting directory: {directory}")
            shutil.rmtree(directory)
            count += 1
    print(f"Deleted {count} directories")


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Delete directories which do not contain a specified subdirectory"
    )
    parser.add_argument(
        "--folder_path", required=True, type=str, help="Path to the folder to clean"
    )
    parser.add_argument(
        "--subdirectory",
        required=True,
        type=str,
        help="Name of the subdirectory to check for in each directory",
    )
    args = parser.parse_args(args)
    delete_dirs_without_subdir(args.folder_path, args.subdirectory)


if __name__ == "__main__":
    main()



import pytest
import torch

from salt.models import InitNet, PositionalEncoder


class MockInputs:
    def __getitem__(self, key):
        if key == "test_input":
            return torch.rand(10, 10, 5)
        if key == "global_object":
            return torch.rand(10, 2)
        raise KeyError


@pytest.fixture
def dense_config():
    return {"hidden_layers": [10, 20], "output_size": 32}


@pytest.fixture
def variables():
    return {"test_input": ["x", "y", "z", "phi", "theta"], "global_object": ["pt", "eta"]}


@pytest.mark.parametrize(
    ("pos_enc", "attach_global"),
    [(None, False), (None, True), (PositionalEncoder(["x", "y"], 32), True)],
)
def test_init_net_forward(dense_config, variables, pos_enc, attach_global):
    input_name = "test_input"
    global_object = "global_object"

    net = InitNet(
        input_name=input_name,
        dense_config=dense_config,
        variables=variables,
        global_object=global_object,
        attach_global=attach_global,
        pos_enc=pos_enc,
    )

    inputs = MockInputs()
    output = net(inputs)

    # Assert output shape is correct
    expected_output_shape = dense_config["output_size"]
    assert output.shape[-1] == expected_output_shape



"""To rapidly train one of the model saved for muP."""

import argparse

from salt.utils.muP_utils.configuration_muP import check_muP, get_models_muPtest

TRAIN = "./pp_output_train.h5"
NORM = "./norm_dict.yaml"
SAVE = "./plots_mup"


def muP_one_model_run(
    modInd, modType, variations, lr, nsteps, nseeds, num_workers, batch_size, num_train
):
    models_muP, models_sP = get_models_muPtest(variations, modInd, modType)
    save = f"{SAVE}_lr_{lr}_batch_{batch_size}"
    check_muP(
        models_muP,
        models_sP,
        lr=lr,
        nsteps=nsteps,
        nseeds=nseeds,
        num_train=num_train,
        batch_size=batch_size,
        num_workers=num_workers,
        train_file=TRAIN,
        norm_dict=NORM,
        save_path=save,
        doPlots=False,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Optional app description")
    parser.add_argument("--modInd", type=int, required=True, help="Ind for the model index")
    parser.add_argument("--modType", type=int, required=True, help="1 for muP, 0 for sP")
    parser.add_argument(
        "--variations", type=int, nargs="+", required=True, help="1 for muP, 0 for sP"
    )
    parser.add_argument("--lr", type=float, default=1e-4, required=False)
    parser.add_argument("--nsteps", type=int, default=2, required=False)
    parser.add_argument("--nseeds", type=int, default=1, required=False)
    parser.add_argument("--num_workers", type=int, default=0, required=False)
    parser.add_argument("--batch_size", type=int, default=500, required=False)
    parser.add_argument("--num_train", type=int, default=-1, required=False)
    args = parser.parse_args()
    muP_one_model_run(
        args.modInd,
        args.modType,
        args.variations,
        args.lr,
        args.nsteps,
        args.nseeds,
        args.num_workers,
        args.batch_size,
        args.num_train,
    )



from torch import nn

from salt.models import Dense, FeaturewiseTransformation
from salt.models.posenc import PositionalEncoder
from salt.stypes import Tensors, Vars
from salt.utils.tensor_utils import attach_context


class InitNet(nn.Module):
    def __init__(
        self,
        input_name: str,
        dense_config: dict,
        variables: Vars,
        global_object: str,
        attach_global: bool = True,
        pos_enc: PositionalEncoder | None = None,
        muP: bool = False,
        featurewise: FeaturewiseTransformation | None = None,
    ):
        """Initial input embedding network.

        This class can handle global input concatenation and positional encoding.

        Parameters
        ----------
        input_name : str
            Name of the input, must match the input types in the data config
        dense_config : dict
            Keyword arguments for [`salt.models.Dense`][salt.models.Dense],
            the dense network producing the initial embedding. The `input_size`
            argument is inferred automatically by the framework
        variables : Vars
            Input variables used in the forward pass, set automatically by the framework
        global_object : str
            Name of the global object, set automatically by the framework
        attach_global : str, optional
            Concatenate global-level inputs with constituent-level inputs before embedding
        pos_enc : PositionalEncoder, optional
            Positional encoder module to use. See
            [`salt.models.PositionalEncoder`][salt.models.PositionalEncoder] for details.
        muP: bool, optional,
            Whether to use the muP parametrisation (impacts initialisation).
        featurewise: FeaturewiseTransformation, optional
            Networks to apply featurewise transformations to inputs, set automatically by
            the framework
        """
        super().__init__()

        # set input size
        if "input_size" not in dense_config:
            dense_config["input_size"] = len(variables[input_name])
            if attach_global and input_name != "EDGE":
                dense_config["input_size"] += len(variables[global_object])
                if not featurewise:
                    dense_config["input_size"] += len(variables.get("PARAMETERS", []))

        self.input_name = input_name
        self.net = Dense(**dense_config)
        self.variables = variables
        self.attach_global = attach_global
        self.global_object = global_object
        self.pos_enc = pos_enc
        self.muP = muP
        if muP:
            self.net.reset_parameters()
        self.featurewise = featurewise

    def forward(self, inputs: Tensors):
        # get the inputs for this init net
        x = inputs[self.input_name]

        # add global features
        if self.attach_global:
            x = attach_context(x, inputs[self.global_object])

        # add parameters if not using featurewise transformations
        if "PARAMETERS" in self.variables and not self.featurewise:
            x = attach_context(x, inputs["PARAMETERS"])

        # apply featurewise transformations
        if self.featurewise is not None:
            x = self.featurewise(inputs, x)

        # inital projection
        x = self.net(x)

        # add positional encoding
        if self.pos_enc:
            this_vars = self.variables[self.input_name]
            input_indices = [this_vars.index(v) for v in self.pos_enc.variables]
            x += self.pos_enc(inputs[self.input_name][..., input_indices])

        return x



from collections.abc import Mapping

import torch
from torch import Tensor, nn, randn

from salt.models.transformer import TransformerCrossAttentionLayer
from salt.stypes import Tensors
from salt.utils.tensor_utils import flatten_tensor_dict, masked_softmax


class Pooling(nn.Module): ...


class GlobalAttentionPooling(Pooling):
    def __init__(self, input_size: int):
        super().__init__()
        self.gate_nn = nn.Linear(input_size, 1)

    def forward(
        self,
        x: dict[str, Tensor] | dict,
        pad_mask: dict | None = None,
    ):
        x_flat = flatten_tensor_dict(x, exclude=["objects"])

        if pad_mask is not None:
            pad_mask = torch.cat(list(pad_mask.values()), dim=1).unsqueeze(-1)

        weights = masked_softmax(self.gate_nn(x_flat), pad_mask, dim=1)
        # add padded track to avoid error in onnx model when there are no tracks in the jet
        weight_pad = torch.zeros((weights.shape[0], 1, weights.shape[2]), device=weights.device)
        x_pad = torch.zeros((x_flat.shape[0], 1, x_flat.shape[2]), device=x_flat.device)
        weights = torch.cat([weights, weight_pad], dim=1)
        x_flat = torch.cat([x_flat, x_pad], dim=1)

        return (x_flat * weights).sum(dim=1)


class BaseCrossAttentionPooling(Pooling):
    def __init__(
        self,
        input_size: int,
        num_layers: int,
        mha_config: Mapping,
        dense_config: Mapping | None = None,
        context_dim: int = 0,
    ):
        super().__init__()
        self.input_size = input_size
        self.num_layers = num_layers
        self.ca_layers = nn.ModuleList([
            TransformerCrossAttentionLayer(input_size, mha_config, dense_config, context_dim)
            for _ in range(num_layers)
        ])
        self.final_norm = nn.LayerNorm(input_size)
        self.class_token = nn.Parameter(randn(1, 1, input_size))

    def expand_class_token(self, x: Tensor | dict):
        if isinstance(x, dict):
            return self.class_token.expand(x[list(x.keys())[0]].shape[0], 1, self.input_size)

        return self.class_token.expand(x.shape[0], 1, self.input_size)


class DictCrossAttentionPooling(BaseCrossAttentionPooling):
    def forward(
        self,
        x: Tensors,
        pad_mask: dict | None = None,
        context: Tensor | None = None,
    ):
        class_token = self.expand_class_token(x)
        for layer in self.ca_layers:
            new_class_token = torch.zeros_like(class_token)
            for input_name in x:
                if input_name == "objects":
                    continue
                new_class_token += layer(
                    class_token,
                    x[input_name],
                    key_value_mask=pad_mask[input_name] if pad_mask else None,
                    context=context,
                )
            class_token = new_class_token

        class_token = self.final_norm(class_token)
        return class_token.squeeze(1)


class TensorCrossAttentionPooling(BaseCrossAttentionPooling):
    def forward(
        self,
        x: Tensors,
        pad_mask: dict | None = None,
        context: Tensor | None = None,
    ):
        x_flat = flatten_tensor_dict(x, exclude=["objects"])
        class_token = self.expand_class_token(x_flat)
        if pad_mask is not None:
            pad_mask = torch.cat(list(pad_mask.values()), dim=1)

        for layer in self.ca_layers:
            class_token = layer(class_token, x_flat, key_value_mask=pad_mask, context=context)

        class_token = self.final_norm(class_token)
        return class_token.squeeze(1)


class NodeQueryGAP(Pooling):
    def __init__(self, input_size: int):
        super().__init__()
        self.gate_nn_1 = nn.Linear(input_size, 1)
        self.gate_nn_2 = nn.Linear(input_size, 1)

    def forward(
        self,
        x: Tensors,
        pad_mask: dict | None = None,
    ):
        # Global Attention Pooling applied to both the decoder kv embeddings and
        # the encoder embeddings.
        x_nodes = flatten_tensor_dict(x, exclude=["objects"])

        if pad_mask is not None:
            pad_mask = torch.cat(list(pad_mask.values()), dim=1).unsqueeze(-1)

        weights = masked_softmax(self.gate_nn_1(x_nodes), pad_mask, dim=1)

        # add padded track to avoid error in onnx model when there are no tracks in the jet
        weight_pad = torch.zeros((weights.shape[0], 1, weights.shape[2]), device=weights.device)
        x_pad = torch.zeros((x_nodes.shape[0], 1, x_nodes.shape[2]), device=x_nodes.device)
        weights = torch.cat([weights, weight_pad], dim=1)
        x_nodes = torch.cat([x_nodes, x_pad], dim=1)
        pooled_nodes = (x_nodes * weights).sum(dim=1)

        # get pooled queries
        emb_queries = x["objects"]["embed"]
        query_pad = torch.zeros(
            (emb_queries.shape[0], 1, emb_queries.shape[2]),
            device=emb_queries.device,
        )
        padded_queries = torch.cat([emb_queries, query_pad], dim=1)
        weights = self.gate_nn_2(padded_queries).softmax(1)
        pooled_queries = (padded_queries * weights).sum(dim=1)

        # concatenate pooled nodes and queries
        return torch.cat([pooled_nodes, pooled_queries], dim=-1)



import math

import numpy as np

from salt.utils.inputs import as_half


def get_dtype_edge(ds, variables) -> np.dtype:
    """Return a dtype based on derived edge variables."""
    if variables is None:
        raise ValueError("Edge variables need to be specified if edge features are to be included")

    req_vars = []
    for variable in variables:
        if variable == "dR":
            req_vars.extend(["eta", "phi"])
        elif variable == "z":
            req_vars.extend(["pt"])
        elif variable == "kt":
            req_vars.extend(["eta", "phi", "pt"])
        elif variable == "isSelfLoop":
            continue
        elif variable == "subjetIndex":
            req_vars.extend(["subjetIndex"])
        else:
            raise ValueError(f"Edge feature {variable} not recognized")

    if "valid" in ds.dtype.names and "valid" not in variables:
        req_vars.append("valid")

    if missing := set(req_vars) - set(ds.dtype.names):
        raise ValueError(
            f"Variables {missing} required for edge feature calculation were not found in dataset"
            f" {ds.name} in file {ds.file.filename}"
        )

    return np.dtype([(n, as_half(x)) for n, x in ds.dtype.descr if n in req_vars])


def get_inputs_edge(batch, variables):
    """Calculate edge features from batch info."""
    ebatch = np.zeros(
        (batch.shape[0], batch.shape[1], batch.shape[1], len(variables)), dtype=np.float32
    )

    with np.errstate(divide="ignore"):
        # calculate useful intermediate quantities
        if "dR" in variables or "kt" in variables:
            dphi = np.expand_dims(batch["phi"].astype(ebatch.dtype), 1).repeat(
                batch.shape[1], 1
            ) - np.expand_dims(batch["phi"].astype(ebatch.dtype), 2).repeat(batch.shape[1], 2)
            dphi -= np.ones_like(dphi) * (dphi > math.pi) * 2 * math.pi
            deta = np.expand_dims(batch["eta"].astype(ebatch.dtype), 1).repeat(
                batch.shape[1], 1
            ) - np.expand_dims(batch["eta"].astype(ebatch.dtype), 2).repeat(batch.shape[1], 2)
        if "kt" in variables or "z" in variables:
            pt_min = np.minimum(
                np.expand_dims(batch["pt"].astype(ebatch.dtype), 1).repeat(batch.shape[1], 1),
                np.expand_dims(batch["pt"].astype(ebatch.dtype), 2).repeat(batch.shape[1], 2),
            )
        # calculate edge feature information and fill batch
        for i, variable in enumerate(variables):
            if variable == "dR":
                ebatch[:, :, :, i] = np.log(np.sqrt(np.square(deta) + np.square(dphi)))
            elif variable == "kt":
                ebatch[:, :, :, i] = np.log(pt_min * np.sqrt(np.square(deta) + np.square(dphi)))
            elif variable == "z":
                pt_sum = np.expand_dims(batch["pt"].astype(ebatch.dtype), 1).repeat(
                    batch.shape[1], 1
                ) + np.expand_dims(batch["pt"].astype(ebatch.dtype), 2).repeat(batch.shape[1], 2)
                ebatch[:, :, :, i] = np.log(pt_min / pt_sum)
            elif variable == "isSelfLoop":
                ebatch[:, :, :, i] = np.identity(batch.shape[1])
            elif variable == "subjetIndex":
                sji1 = np.expand_dims(batch["subjetIndex"].astype(ebatch.dtype), 1).repeat(
                    batch.shape[1], 1
                )
                sji2 = np.expand_dims(batch["subjetIndex"].astype(ebatch.dtype), 2).repeat(
                    batch.shape[1], 2
                )
                ebatch[:, :, :, i] = np.logical_and(np.equal(sji1, sji2), sji1 >= 0)

    return np.nan_to_num(ebatch, nan=0.0, posinf=0.0, neginf=0.0)



import argparse
import json
import warnings
from pathlib import Path

import onnx
import torch
import yaml
from ftag.git_check import check_for_uncommitted_changes, get_git_hash

from salt.models.maskformer import get_maskformer_outputs
from salt.models.task import mask_fill_flattened
from salt.models.transformer_v2 import change_attn_backends
from salt.modelwrapper import ModelWrapper
from salt.onnx.check import compare_outputs
from salt.utils.configs import MaskformerConfig
from salt.utils.get_structured_input_dict import get_structured_input_dict
from salt.utils.union_find import get_node_assignment_jit

torch.manual_seed(42)
# https://gitlab.cern.ch/atlas/athena/-/blob/master/PhysicsAnalysis/JetTagging/FlavorTagDiscriminants/Root/DataPrepUtilities.cxx
TRACK_SELECTIONS = [
    "all",
    "ip3d",
    "dipsLoose202102",
    "r22default",
    "r22loose",
    "dipsTightUpgrade",
    "dipsLooseUpgrade",
]


def parse_args(args):
    parser = argparse.ArgumentParser(
        description="A script to convert a salt model to ONNX.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--ckpt_path",
        type=Path,
        help="Checkpoint path.",
        required=True,
    )
    parser.add_argument(
        "-c",
        "--config",
        type=Path,
        help="Saved training config. If not provided, look in the parent directory of `ckpt_path`.",
        required=False,
    )
    parser.add_argument(
        "-t",
        "--track_selection",
        type=str,
        help="Track selection, must match `trk_select_regexes` in 'DataPrepUtilities.cxx'",
        choices=TRACK_SELECTIONS,
        default="r22default",
    )
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        help="Model name, used in the *_px outputs. Taken from the config if not provided",
        required=False,
    )
    parser.add_argument(
        "-o",
        "--overwrite",
        help="Overwrite existing exported ONNX model.",
        action="store_true",
    )
    parser.add_argument(
        "--tasks",
        type=str,
        nargs="+",
        default=None,
        help=(
            "Space separated list of tasks to include in the ONNX model. "
            "By default all global tasks are included.",
        ),
    )
    parser.add_argument(
        "-mf",
        "--object_name",
    )
    parser.add_argument(
        "--combine_outputs",
        type=str,
        nargs="+",
        default=[],
        help="Space seperated list of items described in 'parse_output_combination'",
    )
    parser.add_argument(
        "-r",
        "--rename",
        type=str,
        nargs="+",
        default=[],
        help="Space seperated list of them form 'oldname1:newname1 oldname2:newname2'",
    )
    parser.add_argument(
        "-f",
        "--force",
        help="Run with uncomitted changes.",
        action="store_true",
    )

    return parser.parse_args(args)


class ONNXModel(ModelWrapper):
    def __init__(
        self,
        onnx_feature_map: list[dict],
        variable_map: dict,
        name: str | None = None,
        tasks_to_output: list[str] | None = None,
        object_name: str | None = None,
        mf_config: dict | None = None,
        combine_outputs: list[tuple] | None = None,
        rename_outputs: dict[str, str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.name = name or self.name
        assert "_" not in self.name, "Model name cannot contain underscores."
        assert "-" not in self.name, "Model name cannot contain dashes."
        for task in self.model.tasks:
            task.model_name = self.name
        self.tasks_to_output = tasks_to_output or []
        if not self.tasks_to_output:
            self.tasks_to_output = [
                t.name for t in self.model.tasks if t.input_name == self.global_object
            ]
            print("No tasks specified, dumping all global tasks: ", self.tasks_to_output)
        else:
            # Need aux tasks to be after global tasks, and in the following order for checks to work
            for t in ["track_origin", "track_vertexing", "track_type"]:
                if t in self.tasks_to_output:
                    self.tasks_to_output.remove(t)
                    self.tasks_to_output.append(t)

        self.combine_outputs = combine_outputs or []
        self.rename_outputs = rename_outputs or {}
        if sum([bool(object_name), bool(mf_config)]) not in {0, 2}:
            raise ValueError("If one of object name or mf config is defined, so must the other.")
        self.object = object_name
        self.mf_config = MaskformerConfig(**mf_config) if mf_config else None
        if self.object and self.mf_config:
            self.object_params = {
                "class_label": self.mf_config.object.class_label,
                "label_map": [f"p{name}" for name in self.mf_config.object.class_names],
            }

        self.feature_map = onnx_feature_map
        self.aux_sequence_object = "tracks"
        self.variable_map = variable_map

        example_input_list = []
        num_tracks = 40
        self.salt_names = []
        self.input_names = []
        self.aux_sequence_index = 1

        for i, feature in enumerate(self.feature_map):
            if feature["name_salt"] == self.global_object:
                example_input_list.append(torch.rand(1, self.input_dims[self.global_object]))
            else:
                example_input_list.append(
                    torch.rand(1, num_tracks, self.input_dims[feature["name_salt"]]).squeeze(0)
                )
            if feature["name_salt"] == self.aux_sequence_object:
                self.aux_sequence_index = i
            self.salt_names.append(feature["name_salt"])
            self.input_names.append(feature["name_athena_out"])
        self.example_input_array = tuple(example_input_list)

    @property
    def global_tasks(self):
        return [t for t in self.model.tasks if t.input_name == self.global_object]

    @property
    def output_names(self) -> list[str]:
        """The output names are a list of strings, one for each output of the model."""
        outputs = []
        for t in self.global_tasks:
            if t.name in self.tasks_to_output:
                outputs += t.output_names

        # Allow renaming and combining of global tasks
        if self.rename_outputs:
            for oldname, newname in self.rename_outputs.items():
                full_oldname = f"{self.name}_{oldname}"
                full_newname = f"{self.name}_{newname}"
                assert full_oldname in outputs, f"Output {oldname} not found in outputs"
                idx = outputs.index(full_oldname)
                outputs[idx] = full_newname

        if self.combine_outputs:
            for output_name, parsed_inputs in self.combine_outputs:
                for _, input_name in parsed_inputs:
                    assert (
                        f"{self.name}_{input_name}" in outputs
                    ), f"Output {input_name} not found in outputs"
                outputs.append(f"{self.name}_{output_name}")

        if "track_origin" in self.tasks_to_output:
            out_name = f"{self.name}_TrackOrigin"
            outputs.append(out_name)

        if "track_vertexing" in self.tasks_to_output:
            out_name = f"{self.name}_VertexIndex"
            outputs.append(out_name)

        if "track_type" in self.tasks_to_output:
            out_name = f"{self.name}_TrackType"
            outputs.append(out_name)
        if self.object:
            regression_task = [
                t for t in self.model.tasks if t.input_name == "objects" and t.name == "regression"
            ]
            assert len(regression_task) == 1, "Object outputs require a regression task"
            # First we append the leading jet regression variables
            outputs += [
                f"{self.name}_leading_{self.object}_{v}" for v in regression_task[0].targets
            ]
            outputs += [f"{self.name}_{self.object}Index"]

        return outputs

    @property
    def dynamic_axes(self) -> dict[str, dict[int, str]]:
        """Let ONNX know which inputs/outputs have dynamic shape (i.e. can vary in length)."""
        # dynamic inputs
        dynamic_axes = {}
        for feature in self.feature_map:
            if not feature["is_global"]:
                dynamic_axes.update({feature["name_athena_out"]: {0: feature["athena_num_name"]}})
        # dynamic outputs
        if "track_origin" in self.tasks_to_output:
            out_name = f"{self.name}_TrackOrigin"
            dynamic_axes[out_name] = {0: "n_tracks"}
        if "track_vertexing" in self.tasks_to_output:
            out_name = f"{self.name}_VertexIndex"
            dynamic_axes[out_name] = {0: "n_tracks"}
        if "track_type" in self.tasks_to_output:
            out_name = f"{self.name}_TrackType"
            dynamic_axes[out_name] = {0: "n_tracks"}
        if self.object:
            out_name = f"{self.name}_{self.object}"
            dynamic_axes[out_name] = {0: "n_tracks"}
        return dynamic_axes

    def forward(self, *args):  # type: ignore[override]
        """Forward pass through the model.

        The arguments must be passed in the same order they appear in the feature map.
        """
        # in athena the jets have a batch dim but the tracks don't, so add it here
        assert len(args) == len(self.salt_names), "Number of inputs does not match feature map."
        assert (
            len(args[0].shape) == 2
        ), "Jets should have a batch dimension, and variable dimension but not a sequence dimension"
        input_dict = {self.global_object: args[0]}
        input_dict.update({
            self.salt_names[i]: args[i].unsqueeze(0) for i in range(1, len(self.salt_names))
        })

        structured_input_dict = get_structured_input_dict(
            input_dict, self.variable_map, self.global_object
        )

        masks = {
            k: torch.zeros((1, args[i].shape[0]), dtype=torch.bool)
            for i, k in enumerate(self.salt_names)
            if k != self.global_object
        }

        outputs = super().forward(input_dict, masks)[0]

        # get the global tasks outputs
        onnx_outputs = sum(
            (
                t.get_onnx(outputs[self.global_object][t.name], labels=structured_input_dict)
                for t in self.global_tasks
            ),
            (),
        )

        if self.combine_outputs:
            onnx_out_names = self.output_names
            for _, parsed_inputs in self.combine_outputs:
                output = sum([
                    scale * onnx_outputs[onnx_out_names.index(f"{self.name}_{input_name}")]
                    for scale, input_name in parsed_inputs
                ])
                onnx_outputs += (output,)

        # add aux outputs
        if "track_origin" in self.tasks_to_output:
            track_origins = outputs[self.aux_sequence_object]["track_origin"]
            track_origins = torch.concatenate(
                [track_origins, torch.zeros((1, 1, track_origins.shape[-1]))], dim=1
            )

            outputs_track = torch.argmax(track_origins, dim=-1)[:, :-1]
            outputs_track = outputs_track.squeeze(0).char()

            onnx_outputs += (outputs_track,)

        if "track_vertexing" in self.tasks_to_output:
            tracks = args[self.aux_sequence_index].unsqueeze(0)
            pad_mask = torch.zeros(tracks.shape[:-1], dtype=torch.bool)
            edge_scores = outputs[self.aux_sequence_object]["track_vertexing"]

            vertex_indices = get_node_assignment_jit(edge_scores, pad_mask)
            vertex_list = mask_fill_flattened(vertex_indices, pad_mask)
            onnx_outputs += (vertex_list.reshape(-1).char(),)

        if "track_type" in self.tasks_to_output:
            track_type = outputs[self.aux_sequence_object]["track_type"]
            track_type = torch.concatenate(
                [track_type, torch.zeros((1, 1, track_type.shape[-1]))], dim=1
            )

            outputs_track = torch.argmax(track_type, dim=-1)[:, :-1]
            outputs_track = outputs_track.squeeze(0).char()
            onnx_outputs += (outputs_track,)

        if self.object:
            assert "objects" in outputs, "No MF objects in outputs"
            regression_tasks = [
                t for t in self.model.tasks if t.input_name == "objects" and t.name == "regression"
            ]
            assert len(regression_tasks) == 1, "Object outputs require a regression task"
            regression_task = regression_tasks[0]

            # Get the (hopefully) correctly (un)scaled regression predictions
            for i, t in enumerate(regression_task.targets):
                unscaled_preds = regression_task.scaler.inverse(
                    t, outputs["objects"]["regression"][:, :, i]
                )
                outputs["objects"]["regression"][:, :, i] = unscaled_preds

            # Extract the mf outputs.
            # TODO: write all regression values, this will require work on the athena end as well
            # https://gitlab.cern.ch/atlas-flavor-tagging-tools/algorithms/salt/-/issues/53
            leading_reg, indices, class_probs, regression = get_maskformer_outputs(  # noqa: F841
                outputs["objects"], apply_reorder=True
            )

            for r in leading_reg[0]:
                onnx_outputs += (r,)
            onnx_outputs += (indices.reshape(-1).char(),)

        return onnx_outputs


def get_default_onnx_feature_map(track_selection, inputs, global_name):
    feature_map = []

    for input_name in inputs:
        if input_name == global_name:
            feature_map.append({
                "name_athena_in": f"{global_name.removesuffix('s')}_var",
                "name_athena_out": f"{global_name.removesuffix('s')}_features",
                "name_salt": global_name,
                "is_global": True,
            })
        elif "tracks" in input_name or "flows" in input_name:
            base_name = input_name.split("_")[0]
            feature_map.append({
                "name_athena_in": f"{base_name}_{track_selection}_sd0sort",
                "name_athena_out": f"{base_name.removesuffix('s')}_features",
                "athena_num_name": f"n_{base_name}",
                "name_salt": base_name,
                "is_global": False,
            })
        else:
            feature_map.append({
                "name_athena_in": f"{input_name}_var",
                "name_athena_out": f"{input_name.removesuffix('s')}_features",
                "athena_num_name": f"n_{input_name}",
                "name_salt": input_name,
                "is_global": False,
            })

    return feature_map


def parse_output_combination(arg_str):
    """Parses a string which represents combinations of outputs to be included in the final
    onnx model.

    Parameters
    ----------
    arg_str : str
        String of the form:
            - light,0.5*q,0.1*g,0.7*s
                Represents p_light=0.5*p_q + 0.1*p_g + 0.7*p_s
            - light,q,g,s
                Represents p_light=p_q + p_g + p_s
        A commar separated list of where the first is the output name, and each other item
        represents a scaled sum

    Returns
    -------
    output_name : str
        The name of the output
    parsed_inputs : list[tuple[float, str]]
        A list of tuples where the first element is the scale and the second is the input name
    """
    all_terms = arg_str.split(",")
    output_name = all_terms[0]
    terms = all_terms[1:]
    parsed_inputs = []
    for term in terms:
        if "*" in term:
            scale, input_name = term.split("*")
            parsed_inputs.append((float(scale), input_name))
        else:
            parsed_inputs.append((1, term))
    return output_name, parsed_inputs


def main(args=None):
    # parse args
    args = parse_args(args)

    if not args.force:
        check_for_uncommitted_changes(Path(__file__).parent)

    if not (config_path := args.config):
        config_path = args.ckpt_path.parents[1] / "config.yaml"
        assert config_path.is_file(), f"Could not find config file at {config_path}"

    config = yaml.safe_load(config_path.read_text())
    # Default config that only uses jets and tracks sorted in a default way
    onnx_feature_map = get_default_onnx_feature_map(
        args.track_selection,
        list(config["data"]["variables"].keys()),
        config["data"]["global_object"],
    )

    combine_outputs = []
    if args.combine_outputs:
        for output in args.combine_outputs:
            output_name, parsed_inputs = parse_output_combination(output)
            combine_outputs.append((output_name, parsed_inputs))

    rename_outputs = dict(x.split(":") for x in args.rename)
    # instantiate pytorch and wrapper models
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        pt_model = ModelWrapper.load_from_checkpoint(
            args.ckpt_path,
            map_location=torch.device("cpu"),
            norm_config=config["model"]["norm_config"],
        )
        pt_model.eval()
        pt_model.float()
        change_attn_backends(pt_model.model, "torch-math")

        if args.object_name:
            with open(config_path) as f:
                config = yaml.safe_load(f)
            mf_config = config["data"].get("mf_config")
            if not mf_config:
                raise ValueError("No mf_config in config")
        else:
            mf_config = {}

        onnx_model = ONNXModel.load_from_checkpoint(
            args.ckpt_path,
            onnx_feature_map=onnx_feature_map,
            variable_map=config["data"]["variables"],
            name=args.name,
            tasks_to_output=args.tasks,
            object_name=args.object_name,
            mf_config=mf_config,
            map_location=torch.device("cpu"),
            norm_config=config["model"]["norm_config"],
            combine_outputs=combine_outputs,
            rename_outputs=rename_outputs,
        )

        onnx_model.eval()
        onnx_model.float()
        change_attn_backends(
            onnx_model.model, "torch-math"
        )  # Only applies to transformer_v2 layers

    print("\n" + "-" * 100)
    print("Converting model to ONNX...")
    print("-" * 100)

    # configure paths
    base_path = args.ckpt_path.parent.parent
    onnx_path = base_path / "network.onnx"
    if onnx_path.exists() and not args.overwrite:
        raise FileExistsError(f"Found existing file '{onnx_path}'.")

    onnx_model.to_onnx(
        onnx_path,
        opset_version=16,
        input_names=onnx_model.input_names,
        output_names=onnx_model.output_names,
        dynamic_axes=onnx_model.dynamic_axes,
    )

    # add metadata
    add_metadata(
        config_path,
        config,
        args.ckpt_path,
        onnx_path,
        onnx_model.name,
        onnx_model.output_names,
        onnx_feature_map,
        combine_outputs,
        rename_outputs,
    )
    seq_names_onnx = []
    seq_names_salt = []
    for feature in onnx_feature_map:
        if feature["is_global"]:
            continue
        seq_names_salt.append(feature["name_salt"])
        seq_names_onnx.append(feature["name_athena_out"])

    # validate pytorch and exported onnx models
    compare_outputs(
        pt_model,
        onnx_path,
        global_object=config["data"]["global_object"],
        seq_names_salt=seq_names_salt,
        seq_names_onnx=seq_names_onnx,
        variable_map=config["data"]["variables"],
        tasks_to_output=onnx_model.tasks_to_output,
    )
    print("\n" + "-" * 100)
    print(f"Done! Saved ONNX model at {onnx_path}")
    print("-" * 100)
    print()


def add_metadata(
    config_path,
    config,
    ckpt_path,
    onnx_path,
    model_name,
    output_names,
    onnx_feature_map,
    combine_outputs,
    rename_outputs,
):
    print("\n" + "-" * 100)
    print("Adding Metadata...")

    # load and check the model
    onnx_model = onnx.load(onnx_path)
    onnx.checker.check_model(onnx_model)

    # add metadata
    metadata = {"ckpt_path": str(ckpt_path.resolve()), "layers": [], "nodes": []}
    metadata["config.yaml"] = config
    metadata["metadata.yaml"] = yaml.safe_load((config_path.parent / "metadata.yaml").read_text())
    metadata["salt_export_hash"] = get_git_hash(Path(__file__).parent)

    # careful - this stuff is used in athena
    metadata["onnx_model_version"] = "v1"
    metadata["output_names"] = output_names
    metadata["model_name"] = model_name
    metadata["inputs"] = []
    metadata["input_sequences"] = []
    metadata["combine_outputs"] = combine_outputs
    metadata["rename_outputs"] = rename_outputs
    for feature in onnx_feature_map:
        if feature["is_global"]:  # global features similar to jet features
            metadata["inputs"] += [
                {
                    "name": feature["name_athena_in"],
                    "variables": [
                        {"name": k.removesuffix("_btagJes"), "offset": 0.0, "scale": 1.0}
                        for k in config["data"]["variables"][feature["name_salt"]]
                    ],
                }
            ]
        else:  # feature sequences simmilar to tracks features
            metadata["input_sequences"] += [
                {
                    "name": feature["name_athena_in"],
                    "variables": [
                        {"name": k, "offset": 0.0, "scale": 1.0}
                        for k in config["data"]["variables"][feature["name_salt"]]
                    ],
                },
            ]

    # write metadata as json string
    metadata = {"gnn_config": json.dumps(metadata)}

    for k, v in metadata.items():
        meta = onnx_model.metadata_props.add()
        meta.key = k
        meta.value = v

    onnx_model.doc_string = model_name
    onnx.save(onnx_model, onnx_path)
    print("-" * 100)


if __name__ == "__main__":
    main()



def get_structured_input_dict(input_dict, variable_map, global_object):
    structured_input_dict = {}
    for obj_type in [global_object]:  # should be extended to include non-global objs ideally
        structured_input_dict[obj_type] = {}
        for i, var in enumerate(variable_map[obj_type]):
            structured_input_dict[obj_type][var] = input_dict[obj_type][:, i]
    return structured_input_dict



import numpy as np


def join_structured_arrays(arrays: list):
    """Join a list of structured numpy arrays.

    See https://github.com/numpy/numpy/issues/7811

    Parameters
    ----------
    arrays : list
        List of structured numpy arrays to join

    Returns
    -------
    np.array
        A merged structured array
    """
    dtype: list = sum((a.dtype.descr for a in arrays), [])
    newrecarray = np.empty(arrays[0].shape, dtype=dtype)
    for a in arrays:
        for name in a.dtype.names:
            newrecarray[name] = a[name]

    return newrecarray


def listify(maybe_list):
    """Convert a scalar or list to a list.
    If None, returns None.

    Parameters
    ----------
    maybe_list
        A scalar, list or None

    Returns
    -------
    list
        A list or None
    """
    if maybe_list is None:
        return None
    if isinstance(maybe_list, list):
        return maybe_list
    return [maybe_list]


def maybe_pad(src: np.ndarray, tgt: np.ndarray) -> np.ndarray:
    """Pad src to the shape of tgt if necessary."""
    if src.shape == tgt.shape:
        return src
    seq_len = tgt.shape[1] if tgt.ndim == 2 else None
    if seq_len and seq_len != src.shape[1]:
        n_pad = seq_len - src.shape[1]
        src = np.pad(src, ((0, 0), (0, n_pad)), mode="constant")
    return src


def maybe_copy(src: np.ndarray):
    """Return a copy of src if it is not C-contiguous."""
    if src.flags.c_contiguous:
        return src
    return src.copy()



"""Repair checkpoint files after compilation.

Adapted from:
https://github.com/pytorch/pytorch/issues/101107#issuecomment-1801128683
"""

import argparse
import shutil
from pathlib import Path

import torch


def repair_checkpoint(path: str | Path) -> None:
    """Repair a PyTorch checkpoint file by removing specific prefixes from state_dict keys.

    Parameters
    ----------
    path : str
        The path to the checkpoint file to repair.
    """
    path = Path(path)
    ckpt = torch.load(path)
    in_state_dict = ckpt["state_dict"]
    pairings = [(src_key, src_key.replace("_orig_mod.", "")) for src_key in in_state_dict]

    if all(src_key == dest_key for src_key, dest_key in pairings):
        print(f"No need to repair {path}")
        return

    shutil.copyfile(path, str(path) + ".bak")
    print(f"Backup created: {path}.bak")

    out_state_dict = {}
    for src_key, dest_key in pairings:
        print(f"{src_key}  ==>  {dest_key}")
        out_state_dict[dest_key] = in_state_dict[src_key]

    ckpt["state_dict"] = out_state_dict
    torch.save(ckpt, path)
    print(f"Repaired {path}")


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Repair PyTorch checkpoint files trained with torch.compile()."
    )
    parser.add_argument("paths", nargs="+", help="The file paths of the checkpoints to repair.")
    args = parser.parse_args(args)
    for path in args.paths:
        repair_checkpoint(path)


if __name__ == "__main__":
    main()



import math
import warnings
from collections.abc import Mapping

import lightning as L
import torch
from torch import nn

from salt.models import InputNorm


def check_unique(modules: nn.ModuleList, attr_name: str) -> None:
    assert len({getattr(m, attr_name) for m in modules}) == len(
        modules
    ), f"Attribute '{attr_name}' must be unique for class {modules[0].__class__.__name__}"


class ModelWrapper(L.LightningModule):
    def __init__(
        self,
        model: nn.Module,
        lrs_config: Mapping[str, float],
        global_object: str,
        norm_config: dict | None = None,
        name: str = "salt",
        muP_config: dict | None = None,
        loss_mode: str = "wsum",
        optimizer: str = "AdamW",
    ):
        """A wrapper class for any model implemented in Salt.

        This wrapper class allows is as generic as possible. It wraps
        [`SaltModel`][salt.models.SaltModel], but could also be used to
        wrap any other model if you want to do train something that doesn't
        fit into the [`SaltModel`][salt.models.SaltModel] architecture.

        This class is responsible for containing things that are common to all
        salt models. These are:

        - A generic forward pass, including input normalisation
        - Training, validation and test steps, which include logging
        - Some sanity checks on the model configuration

        Parameters
        ----------
        model : nn.Module
            Model to be wrapped
        lrs_config: Mapping
            LRS config which has to be set manually for now
            https://github.com/omni-us/jsonargparse/issues/170#issuecomment-1288167674
        global_object : str
            Name of the global input object, as opposed to the constituent-level
            inputs. This argument is set automatically by the framework.
        norm_config : dict, optional
            Keyword arguments for [`salt.models.InputNorm`][salt.models.InputNorm].
        name: str, optional
            Name of the model, used for logging and inference output names
        muP_config: dict, optional
            The muP configuration.
        loss_mode: str, optional
            The loss mode to use. Default is "wsum" (weighted sum).
            Other options are
            - 'GLS' : arxiv.org/1904.08492
        optimizer: str, optional
            Optimizer used. Default if "AdamW"
            Other options are
            - 'lion': https://github.com/lucidrains/lion-pytorch
        """
        super().__init__()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.save_hyperparameters(logger=False)

        self.model = model
        self.lrs_config = lrs_config
        self.global_object = global_object
        self.name = name
        self.muP = muP_config or {}
        self.last_val_batch_outs = None
        # Here the model should pick it up
        if self.muP:
            from salt.utils.muP_utils.configuration_muP import instantiate_mup

            load_path = None
            if "shape_path" in self.muP:
                load_path = self.muP["shape_path"]
            instantiate_mup(model, load_path)

        # all tasks should inherit the global object type and model name
        self.model.global_object = self.global_object
        for task in self.model.tasks:
            task.global_object = self.global_object
            task.model_name = self.name

        # ensure unique names for init_nets and tasks
        check_unique(self.model.init_nets, "input_name")
        check_unique(self.model.tasks, "name")

        # check that the model has the same output size for all init nets
        assert len({t.net.output_size for t in self.model.init_nets if t.input_name != "EDGE"}) == 1

        # create input normaliser
        assert norm_config is not None
        self.norm = InputNorm(**norm_config)
        allowed_loss_modes = ["wsum", "GLS"]
        assert loss_mode in allowed_loss_modes, f"Loss mode must be one of {allowed_loss_modes}"
        self.loss_mode = loss_mode
        if loss_mode == "GLS":
            assert all(
                task.weight == 1.0 for task in self.model.tasks
            ), "GLS does not utilise task weights - remove all/set to 1"
        allowed_optimizers = ["lion", "AdamW"]
        assert (
            optimizer in allowed_optimizers
        ), f"Optimizer {optimizer} not implemented, please choose from {allowed_optimizers}"
        self.optimizer = optimizer

    def total_loss(self, loss: dict):
        """Computes the final loss based on the loss mode."""
        if self.loss_mode == "GLS":
            # Calculate the geometric mean of the losses
            loss_prod = math.prod(subloss for subloss in loss.values())
            return torch.pow(loss_prod, 1.0 / len(loss))

        # Return the default weighted sum
        return sum(subloss for subloss in loss.values())

    def forward(self, inputs, pad_masks=None, labels=None):
        """Generic forward pass through any salt-compatible model.

        This function performs input normalisation and then calls the `self.model`'s
        forward pass. Don't call this method directy, instead use `__call__`.

        Parameters
        ----------
        inputs
            Any generic input to the model.
        pad_masks
            Input padding masks.
        labels
            Training targets. If not specified, assume we are running model inference
            (i.e. no loss computation).

        Returns
        -------
        Whatever is returned by `self.model`'s forward pass.
        """
        x = self.norm(inputs)
        return self.model(x, pad_masks, labels)

    def shared_step(self, batch, evaluation=False):
        """Function used to unpack the batch, run the forward pass, and compute
        losses, used by training, validation and test steps.

        Parameters
        ----------
        batch : tuple
            A single batch of inputs, pad_masks and labels
        evaluation : bool
            If true, don't compute the losses and return early

        Returns
        -------
        preds
            Model predictions
        labels
            True labels
        loss
            Reduced loss over the input batch
        """
        # unpack the batch
        inputs, pad_masks, labels = batch

        # forward pass through model
        preds, loss = self(inputs, pad_masks, labels)

        if evaluation:
            return preds, labels, pad_masks, None

        # compute total loss
        loss["loss"] = self.total_loss(loss)

        return preds, labels, pad_masks, loss

    def log_losses(self, loss, stage):
        kwargs = {"sync_dist": len(self.trainer.device_ids) > 1}
        self.log(f"{stage}/loss", loss["loss"], **kwargs)
        for t, loss_value in loss.items():
            n = f"{stage}/{t}_loss" if "loss" not in t else f"{stage}/{t}"
            self.log(n, loss_value, **kwargs)

    def training_step(self, batch):
        # foward pass
        preds, labels, pad_masks, loss = self.shared_step(batch)

        if loss["loss"].isnan():
            raise RuntimeError(
                "Loss is NaN - this indicates something significant has gone wrong."
                "Check for any NaNs or infs in the input dataset. If nothing is found here, "
                "check 'docs/training.md - NaNs' for more information"
            )

        # log losses
        self.log_losses(loss, stage="train")

        outputs = {
            "preds": preds,
            "labels": labels,
            "pad_masks": pad_masks,
        }

        return {**loss, "outputs": outputs}

    def validation_step(self, batch):
        # foward pass
        preds, labels, pad_masks, loss = self.shared_step(batch)

        # log losses
        self.log_losses(loss, stage="val")

        # Store outputs to be used by the MaskformerMetrics callback
        outputs = {
            "preds": preds,
            "labels": labels,
            "pad_masks": pad_masks,
        }

        return {**loss, "outputs": outputs}

    def test_step(self, batch):
        if (
            type(self.model.encoder).__name__ == "TransformerV2"
            and self.trainer.precision == "32-true"
        ):
            from salt.models.transformer_v2 import change_attn_backends

            change_attn_backends(self, backend="torch-math")
        inputs, pad_masks, _ = batch
        batch = (inputs, pad_masks, None)
        return self.shared_step(batch, evaluation=True)[0]

    def configure_optimizers(self):
        if self.optimizer == "lion":
            from lion_pytorch import Lion

            opt = Lion(
                self.parameters(),
                lr=self.lrs_config["initial"],
                weight_decay=self.lrs_config.get("weight_decay", 1e-5),
            )
        elif self.optimizer == "AdamW":
            if self.muP:
                from mup import MuAdamW as AdamW
            else:
                from torch.optim import AdamW
            opt = AdamW(
                self.parameters(),
                lr=self.lrs_config["initial"],
                weight_decay=self.lrs_config.get("weight_decay", 1e-5),
            )

        # 1cycle
        sch = torch.optim.lr_scheduler.OneCycleLR(
            opt,
            max_lr=self.lrs_config["max"],
            total_steps=self.trainer.estimated_stepping_batches,
            div_factor=self.lrs_config["max"] / self.lrs_config["initial"],
            final_div_factor=self.lrs_config["initial"] / self.lrs_config["end"],
            pct_start=float(self.lrs_config["pct_start"]),
            last_epoch=int(self.lrs_config.get("last_epoch", -1)),
        )
        sch = {"scheduler": sch, "interval": "step"}

        return [opt], [sch]

    @property
    def input_dims(self) -> dict[str, int]:
        return {k: len(v) for k, v in self.norm.variables.items()}



"""Benchmarking utilities for Pytorch models."""

from collections.abc import Callable

import torch
from torch import Tensor, dtype
from torch.utils import benchmark


def time_forward(
    fn: Callable,
    *args,
    repeats: int = 10,
    block_time: float = 0.0,
    desc: str = "",
    verbose: bool = True,
    amp: bool = False,
    amp_dtype: dtype = torch.float16,
    **kwargs,
) -> tuple:
    """Use Pytorch Benchmark on the forward pass of an arbitrary function.

    Parameters
    ----------
    fn : function
        The function to benchmark.
    args : list
        The args to the function.
    repeats : int
        Number of times to repeat the benchmark.
    block_time : float
        Instead of repeats, run the benchmark for a fixed amount of time.
    desc : str
        Description of the benchmark.
    verbose : bool
        Whether to print the benchmark results.
    amp : bool
        Whether to use automatic mixed precision.
    amp_dtype : torch.dtype
        The dtype to use for automatic mixed precision.
    kwargs : dict
        Additional keyword arguments to pass to the function.
    """
    if verbose:
        print(desc, " - Foward pass")

    # Define the automatic mixed precision wrapper
    def fn_with_amp(*args, **kwargs):
        with torch.autocast(device_type="cuda", dtype=amp_dtype, enabled=amp):
            fn(*args, **kwargs)

    # Create the benchmark timer
    t = benchmark.Timer(
        stmt="fn_with_amp(*args, **kwargs)",
        globals={"fn_with_amp": fn_with_amp, "args": args, "kwargs": kwargs},
        num_threads=torch.get_num_threads(),
    )

    # Run the benchmark
    m = t.blocked_autorange(min_run_time=block_time) if block_time > 0 else t.timeit(repeats)

    if verbose:
        print(m)
    return t, m


def time_backward(
    fn: Callable,
    *args,
    repeats: int = 10,
    block_time: float = 0.0,
    desc: str = "",
    verbose: bool = True,
    amp: bool = False,
    amp_dtype: dtype = torch.float16,
    **kwargs,
) -> tuple:
    """Use Pytorch Benchmark on the backward pass of an arbitrary function.

    Parameters
    ----------
    fn : function
        The function to benchmark.
    args : list
        The args to the function.
    repeats : int
        Number of times to repeat the benchmark.
    block_time : float
        Instead of repeats, run the benchmark for a fixed amount of time.
    desc : str
        Description of the benchmark.
    verbose : bool
        Whether to print the benchmark results.
    amp : bool
        Whether to use automatic mixed precision.
    amp_dtype : torch.dtype
        The dtype to use for automatic mixed precision.
    kwargs : dict
        Additional keyword arguments to pass to the function.
    """
    if verbose:
        print(desc, " - Backward pass")

    # Run in forward to get the output so we can backpropagate
    with torch.autocast(device_type="cuda", dtype=amp_dtype, enabled=amp):
        y = fn(*args, **kwargs)
        if type(y) is tuple:
            y = y[0]
        elif type(y) is dict:
            y = next(iter(y.values()))

    # Generate a random gradient
    grad = torch.randn_like(y)

    # Define the backward function
    def bwd(*args, y, grad):
        for x in args:  # Turn off gradients for all args
            if isinstance(x, Tensor):
                x.grad = None
        y.backward(grad, retain_graph=True)

    # Create the benchmark timer
    t = benchmark.Timer(
        stmt="f(*args, y=y, grad=grad)",
        globals={"f": bwd, "args": args, "y": y, "grad": grad},
        num_threads=torch.get_num_threads(),
    )

    # Run the benchmark
    m = t.blocked_autorange(min_run_time=block_time) if block_time > 0 else t.timeit(repeats)
    if verbose:
        print(m)
    return t, m


def benchmark_gpu_memory(
    fn: Callable,
    *args,
    amp: bool = False,
    amp_dtype: dtype = torch.float16,
    **kwargs,
) -> tuple:
    """Calculate the maximum GPU memory used by a function.

    Parameters
    ----------
    fn : function
        The function to benchmark.
    args : list
        The args to the function.
    amp : bool
        Whether to use automatic mixed precision.
    amp_dtype : torch.dtype
        The dtype to use for automatic mixed precision.
    kwargs : dict
        Additional keyword arguments to pass to the function.
    """
    # Clear the cache and reset memory stats
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()
    torch.cuda.synchronize()

    # Run the function
    with torch.autocast(device_type="cuda", dtype=amp_dtype, enabled=amp):
        fn(*args, **kwargs)
    torch.cuda.synchronize()

    # Calculate the max memory used in GB
    mem = torch.cuda.max_memory_allocated() / ((2**20) * 1000)
    torch.cuda.empty_cache()
    return mem



import torch


class RegressionTargetScaler:
    def __init__(self, scales):
        """Generates a functional based scaler for regression targets.
        scales should be a dict of the following form:
        scales = {
            'pt': {
                'op' : 'log',
                'x_scale' : 5,
                'x_off' : 0,
                'op_scale' : 0,
                'op_off' : 0
            },
            'Lxy': {
                'op' : 'log',
                'x_scale' : 5,
                'x_off' : 0,
                'op_scale' : 0,
                'op_off' : 0
            },
            'deta' : {
                'op' : 'linear',
                'x_scale' : 1,
                'x_off' : 1,
                'op_scale' : 10,
                'op_off' : 0
            }
        }
        This would result in scaling functions of the form:
            func = lambda x: (op(x_scale * x + x_off) * op_scale + op_off)
            scale(pt) = lambda x: (log(5 * x) * 0 + 0)
            scale(Lxy) = lambda x: (log(5 * x) * 0 + 0)
            scale(deta) = lambda x: (linear(1 * x + 1) * 10 + 0)
        Only the operation is required, the rest is optional, with defaults
        of 1 for scales and 0 for offsets.
        """
        self.scales = scales

    @staticmethod
    def log_scale(x: torch.Tensor, x_scale=1, x_off=0, op_scale=1, op_off=0) -> torch.Tensor:
        return torch.log(x * x_scale + x_off) * op_scale + op_off

    @staticmethod
    def log_inverse(x: torch.Tensor, x_scale=1, x_off=0, op_scale=1, op_off=0) -> torch.Tensor:
        return (torch.exp((x - op_off) / op_scale) - x_off) / x_scale

    @staticmethod
    def exp_scale(x: torch.Tensor, x_scale=1, x_off=0, op_scale=1, op_off=0) -> torch.Tensor:
        return torch.exp(x * x_scale + x_off) * op_scale + op_off

    @staticmethod
    def exp_inverse(x: torch.Tensor, x_scale=1, x_off=0, op_scale=1, op_off=0) -> torch.Tensor:
        return (torch.log((x - op_off) / op_scale) - x_off) / x_scale

    @staticmethod
    def linear_scale(x: torch.Tensor, x_scale=1, x_off=0, op_scale=1, op_off=0) -> torch.Tensor:
        return (x * x_scale + x_off) * op_scale + op_off

    @staticmethod
    def linear_inverse(x: torch.Tensor, x_scale=1, x_off=0, op_scale=1, op_off=0) -> torch.Tensor:
        return ((x - op_off) / op_scale - x_off) / x_scale

    def scale(self, target: str, values: torch.Tensor) -> torch.Tensor:
        params = self.scales[target].copy()
        op = params.pop("op")
        if op == "log":
            return self.log_scale(values, **params)
        if op == "exp":
            return self.exp_scale(values, **params)
        if op == "linear":
            return self.linear_scale(values, **params)

        raise ValueError("Unknown operation: {}".format(params["op"]))

    def inverse(self, target: str, values: torch.Tensor) -> torch.Tensor:
        params = self.scales[target].copy()
        op = params.pop("op")
        if op == "log":
            return self.log_inverse(values, **params)
        if op == "exp":
            return self.exp_inverse(values, **params)
        if op == "linear":
            return self.linear_inverse(values, **params)

        raise ValueError("Unknown operation: {}".format(params["op"]))



import torch
from torch import Tensor, nn

from salt.models import Dense
from salt.stypes import Vars


class FeaturewiseTransformation(nn.Module):
    def __init__(
        self,
        layer: str,
        variables: Vars,
        dense_config_scale: dict | None = None,
        dense_config_bias: dict | None = None,
        apply_norm: bool = False,
    ):
        """Perform feature wise transformations on the features of a layer.
        https://distill.pub/2018/feature-wise-transformations/.

        Parameters
        ----------
        layer : str
            layer to scale/bias (either "input", "encoder", or "global")
        variables : Vars
            Input variables used in the forward pass, set automatically by the framework
        dense_config_scale : dict
            Keyword arguments for [salt.models.Dense][salt.models.Dense],
            the dense network performing the scaling.
        dense_config_bias : dict
            Keyword arguments for [salt.models.Dense][salt.models.Dense],
            the dense network performing the biasing.
        apply_norm : bool
            Apply layer normalisation to the transformed features. By default false.
        """
        super().__init__()

        self.layer = layer
        if layer not in {"input", "encoder", "global"}:
            raise ValueError(
                "Select either 'input', 'encoder' or 'global' layers for featurewise nets."
            )

        self.scale_net = None
        self.bias_net = None
        self.num_features = None
        self.norm = None

        if dense_config_scale:
            dense_config_scale["input_size"] = len(variables.get("PARAMETERS", []))
            self.scale_net = Dense(**dense_config_scale)
            self.num_features = self.scale_net.output_size
        if dense_config_bias:
            dense_config_bias["input_size"] = len(variables.get("PARAMETERS", []))
            self.bias_net = Dense(**dense_config_bias)
            self.num_features = self.bias_net.output_size

        if not self.bias_net and not self.scale_net:
            raise ValueError(
                "Need to specify at least one dense_config_scale or dense_config_bias."
            )

        if apply_norm:
            self.norm = nn.LayerNorm(self.num_features)

    def forward(self, inputs: dict, features: Tensor):
        if "PARAMETERS" not in inputs:
            raise ValueError("Featurewise transformations require 'PARAMETERS'.")
        x = inputs["PARAMETERS"]
        if self.scale_net:
            features = self.scale_net(x).unsqueeze(1) * features
        if self.bias_net:
            features = torch.add(features, self.bias_net(x).unsqueeze(1))
        if self.norm:
            features = self.norm(features)
        return features



import warnings
from collections.abc import Callable
from copy import deepcopy

import h5py
import numpy as np
import torch
from ftag import Cuts
from ftag.track_selector import TrackSelector
from numpy.lib.recfunctions import structured_to_unstructured as s2u
from torch.utils.data import Dataset

from salt.data.edge_features import get_dtype_edge, get_inputs_edge
from salt.stypes import Vars
from salt.utils.array_utils import maybe_copy
from salt.utils.configs import MaskformerConfig
from salt.utils.inputs import as_half
from salt.utils.mask_utils import build_target_masks


class SaltDataset(Dataset):
    def __init__(
        self,
        filename: str,
        norm_dict: str,
        variables: Vars,
        stage: str,
        num: int = -1,
        labels: Vars = None,
        mf_config: MaskformerConfig | None = None,
        input_map: dict[str, str] | None = None,
        num_inputs: dict | None = None,
        non_finite_to_num: bool = False,
        global_object: str = "jets",
        PARAMETERS: dict | None = None,
        selections: dict[str, list[str]] | None = None,
        ignore_finite_checks: bool = False,
        transforms: list[Callable] | None = None,
    ):
        """An efficient map-style dataset for loading data from an H5 file containing structured
        arrays.

        Parameters
        ----------
        filename : str
            Input h5 filepath containing structured arrays
        norm_dict : str
            Path to file containing normalisation parameters
        variables : Vars
            Input variables used in the forward pass for each input type
        stage : str
            Stage of the training process
        num : int, optional
            Number of input samples to use. If `-1`, use all input samples
        labels : Vars
            List of required labels for each input type
        mf_config : MaskformerConfig, optional
            Config for Maskformer matching, by default None
        input_map : dict, optional
            Map names to the corresponding dataset names in the input h5 file.
            If not provided, the input names will be used as the dataset names.
        num_inputs : dict, optional
            Truncate the number of constituent inputs to this number, to speed up training
        non_finite_to_num : bool, optional
            Convert nans and infs to zeros when loading inputs
        global_object : str
            Name of the global input object, as opposed to the constituent-level
            inputs
        PARAMETERS: dict
            Variables used to parameterise the network, by default None.
        selections : dict, optional
            Selections to apply to the input data, by default None.
        ignore_finite_checks: bool, optional
            Ignoring check for non-finite inputs
        transforms: list, optional
            Transformations to apply to the data, by default None.
        """
        super().__init__()
        # check labels have been configured
        self.labels = labels if labels is not None else {}

        # default input mapping: use input names as dataset names
        # allow only partial maps to be provided
        if input_map is None:
            input_map = {k: k for k in variables}

        if "GLOBAL" in input_map:
            input_map["GLOBAL"] = global_object

        if "PARAMETERS" in input_map:
            input_map["PARAMETERS"] = global_object

        self.input_map = input_map
        self.filename = filename
        self.file = h5py.File(self.filename, "r")
        self.num_inputs = num_inputs
        self.non_finite_to_num = non_finite_to_num
        self.global_object = global_object
        self.selections = selections
        self.selectors = {}
        self.transforms = transforms
        if self.selections:
            for key, value in self.selections.items():
                self.selectors[key] = TrackSelector(Cuts.from_list(value))

        # If MaskFormer matching is enabled, extract the relevent labels
        self.mf_config = deepcopy(mf_config)
        if self.mf_config:
            self.input_map["objects"] = self.mf_config.object.name

        self.variables = variables
        self.norm_dict = norm_dict
        self.PARAMETERS = PARAMETERS
        self.stage = stage
        self.rng = np.random.default_rng()

        # check that num_inputs contains valid keys
        if self.num_inputs is not None and not set(self.num_inputs).issubset(self.variables):
            raise ValueError(
                f"num_inputs keys {self.num_inputs.keys()} must be a subset of input variables"
                f" {self.variables.keys()}"
            )

        self.check_file()

        self.input_variables = variables
        assert self.input_variables is not None

        # check parameters listed in variables appear in the same order in the PARAMETERS block
        if "PARAMETERS" in self.input_variables:
            assert self.PARAMETERS is not None
            assert self.input_variables["PARAMETERS"] is not None
            assert len(self.input_variables["PARAMETERS"]) == len(self.PARAMETERS)
            for idx, param_key in enumerate(self.PARAMETERS.keys()):
                assert self.input_variables["PARAMETERS"][idx] == param_key

        # set number of objects
        self.num = self.get_num(num)
        self.ignore_finite_checks = ignore_finite_checks

        self._is_setup = False

    def _setup(self):
        """Setup the dataset."""
        # setup datasets and accessor arrays
        self.dss = {}
        self.arrays = {}
        file = h5py.File(self.filename, "r")

        for internal, external in self.input_map.items():
            self.dss[internal] = file[external]
            this_vars = self.labels[internal].copy() if internal in self.labels else []
            this_vars += self.input_variables.get(internal, [])
            if internal == "EDGE":
                dtype = get_dtype_edge(file[external], this_vars)
            else:
                dtype = get_dtype(file[external], this_vars)
            self.arrays[internal] = np.array(0, dtype=dtype)
        if self.global_object not in self.dss:
            self.dss[self.global_object] = file[self.global_object]

        self._is_setup = True

    def __len__(self):
        """Return the number of samples in the dataset."""
        return int(self.num)

    def __getitem__(self, object_idx):
        """Return on sample or batch from the dataset.

        Parameters
        ----------
        object_idx
            A numpy slice corresponding to a batch of objects.

        Returns
        -------
        tuple
            Dict of tensor for each of the inputs, pad_masks, and labels.
            Each tensor will contain a batch of samples.
        """
        inputs = {}
        labels = {}
        pad_masks = {}

        # Some weird bug seem to occur due to hdf5 not being thread safe. The below ensures
        # that each worker has its own copy of the file to prevent
        if not self._is_setup:
            self._setup()
        # loop over input types
        for input_name in self.input_map:
            # load data (inputs + labels) for this input type
            batch = self.arrays[input_name]
            shape = (object_idx.stop - object_idx.start,) + self.dss[input_name].shape[1:]
            batch.resize(shape, refcheck=False)
            self.dss[input_name].read_direct(batch, object_idx)

            # apply selections for constituent inputs
            if self.selectors and (selector := self.selectors.get(input_name)):
                batch = selector(batch)

            # truncate constituent inputs
            if self.num_inputs is not None and input_name in self.num_inputs:
                assert int(self.num_inputs[input_name]) <= batch.shape[1]
                batch = batch[:, : int(self.num_inputs[input_name])]

            # load edge inputs for this input type
            if input_name == "EDGE":
                inputs[input_name] = torch.from_numpy(
                    get_inputs_edge(batch, self.input_variables[input_name])
                )

            # load PARAMETERS for this input type
            elif input_name == "PARAMETERS":
                flat_array = s2u(batch[self.input_variables[input_name]], dtype=np.float32)

                for ind, param in enumerate(self.PARAMETERS):
                    if self.stage == "fit":
                        # assign random values to inputs with parameters not set to those in the
                        # train list, values are chosen at random from those in the train list
                        # according to probabilities if given, else with equal probability
                        try:
                            prob = self.PARAMETERS[param]["prob"]
                        except KeyError:
                            prob = None
                        mask = ~np.isin(flat_array[:, ind], self.PARAMETERS[param]["train"])
                        random = self.rng.choice(
                            self.PARAMETERS[param]["train"], size=np.sum(mask), p=prob
                        )
                        flat_array[mask, ind] = random

                    if self.stage == "test":
                        # assign parameter values for all objects passed in the 'test' option
                        test_arr = np.full(np.shape(flat_array)[0], self.PARAMETERS[param]["test"])
                        flat_array[:, ind] = test_arr

                inputs[input_name] = torch.from_numpy(flat_array)

            # load standard inputs for this input type
            elif self.input_variables.get(input_name):
                # Apply transforms while variable names are still accessible
                struct_array = batch[self.input_variables[input_name]]
                if self.transforms:
                    for transform in self.transforms:
                        struct_array = transform(struct_array, input_name)

                flat_array = s2u(struct_array, dtype=np.float32)

                if self.non_finite_to_num:
                    flat_array = np.nan_to_num(flat_array, posinf=0, neginf=0)
                inputs[input_name] = torch.from_numpy(maybe_copy(flat_array))

                # apply the input padding mask
                if "valid" in batch.dtype.names and input_name not in {
                    "EDGE",
                    "PARAMETERS",
                    self.global_object,
                    "GLOBAL",
                }:
                    pad_masks[input_name] = ~torch.from_numpy(batch["valid"])
                    inputs[input_name][pad_masks[input_name]] = 0

                # check inputs are finite
                if not torch.isfinite(inputs[input_name]).all():
                    if self.ignore_finite_checks:
                        warnings.warn(
                            f"Non-finite inputs for '{input_name}' in {self.filename}."
                            "But ignore finite flag is on, make sure this is intentional.",
                            stacklevel=2,
                        )
                    else:
                        raise ValueError(
                            f"Non-finite inputs for '{input_name}' in {self.filename}."
                        )

            # process labels for this input type
            if input_name in self.labels:
                labels[input_name] = {}
                for label in self.labels[input_name]:
                    dtype = torch.long if np.issubdtype(batch[label].dtype, np.integer) else None
                    batch_label = maybe_copy(batch[label])
                    labels[input_name][label] = torch.as_tensor(batch_label, dtype=dtype)
                    x = torch.as_tensor(batch_label, dtype=dtype)
                    if input_name == "objects" and label == self.mf_config.object.class_label:
                        for k, v in self.mf_config.object.class_map.items():
                            x[x == k] = v
                            labels[input_name]["object_class"] = x
                    labels[input_name][label] = x

                # hack to handle the old umami train file format
                if input_name == self.global_object and "/" in self.labels:
                    if self.global_object not in labels:
                        labels[self.global_object] = {}
                    for label in self.labels["/"]:
                        labels[input_name][label] = torch.as_tensor(
                            self.file["labels"][object_idx], dtype=torch.long
                        )

        if self.mf_config:
            labels["objects"]["masks"] = build_target_masks(
                labels["objects"][self.mf_config.object.id_label],
                labels[self.mf_config.constituent.name][self.mf_config.constituent.id_label],
            )

        return inputs, pad_masks, labels

    def get_num(self, num_requested: int):
        num_available = len(self.file[self.global_object])

        # not enough objects
        if num_requested > num_available:
            raise ValueError(
                f"Requested {num_requested:,} from {self.global_object}, but only"
                f" {num_available:,} are available in the file {self.filename}."
            )

        # use all objects
        if num_requested < 0:
            return num_available

        # use requested number of objects
        return num_requested

    def check_file(self):
        keys = {self.input_map[k] for k in self.variables}
        available = set(self.file.keys())
        if missing := keys - available - {"EDGE", "GLOBAL", "PARAMETERS"}:
            raise KeyError(
                f"Input file '{self.filename}' does not contain keys {missing}."
                f" Available keys: {available}"
            )
        for k, v in self.variables.items():
            if k == "EDGE":
                continue
            if k == "PARAMETERS":
                continue
            if k == "GLOBAL":
                k = self.global_object  # noqa: PLW2901
            name = self.input_map[k]
            if not isinstance(self.file[name], h5py.Dataset):
                raise KeyError(f"The object '{name}' in file '{self.filename}' is not a dataset.")
            if missing := set(v) - set(self.file[name].dtype.names):
                raise KeyError(
                    f"Variables {missing} are missing from dataset '{name}' in input file"
                    f" '{self.filename}'."
                )


def get_dtype(ds, variables=None) -> np.dtype:
    """Return a dtype based on an existing dataset and requested variables."""
    if variables is None:
        variables = ds.dtype.names
    if "valid" in ds.dtype.names and "valid" not in variables:
        variables.append("valid")

    variables_flat = []
    for item in variables:
        if isinstance(item, list):
            variables_flat += item
        else:
            variables_flat.append(item)

    return np.dtype([(n, as_half(x)) for n, x in ds.dtype.descr if n in variables])



import torch

from salt.utils.union_find import get_node_assignment


def test_no_vertex() -> None:
    """Test union find on a 2 track jet with no vertices."""
    labels = torch.tensor([0, 1])
    pairwise_probs = torch.tensor([[-1.0, -1.0]])
    mask = torch.tensor([[False, False]])
    assert torch.equal(
        get_node_assignment(pairwise_probs.flatten().unsqueeze(dim=-1), mask),
        labels.unsqueeze(dim=-1),
    )


def test_single_vertex_1() -> None:
    """Test union find on a 3 track jet with one two track vertex."""
    labels = torch.tensor([0, 1, 0])
    pairwise_probs = torch.tensor([[-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
    mask = torch.tensor([[False, False, False]])
    assert torch.equal(
        get_node_assignment(pairwise_probs.flatten().unsqueeze(dim=-1), mask),
        labels.unsqueeze(dim=-1),
    )


def test_single_vertex_2() -> None:
    """Test union find on a 3 track jet with one three track vertex."""
    labels = torch.tensor([0, 0, 0])
    pairwise_probs = torch.tensor([[1.0, -1.0], [1.0, 1.0], [-1.0, 1.0]])
    mask = torch.tensor([[False, False, False]])
    assert torch.equal(
        get_node_assignment(pairwise_probs.flatten().unsqueeze(dim=-1), mask),
        labels.unsqueeze(dim=-1),
    )


def test_single_vertex_3() -> None:
    """Test union find on a 4 track jet with one three track vertex."""
    labels = torch.tensor([0, 1, 1, 1])
    pairwise_probs = torch.tensor([
        [-1.0, -1.0, -1.0],
        [-1.0, 1.0, 1.0],
        [-1.0, 1.0, 1.0],
        [-1.0, 1.0, 1.0],
    ])
    mask = torch.tensor([[False, False, False, False]])
    assert torch.equal(
        get_node_assignment(pairwise_probs.flatten().unsqueeze(dim=-1), mask),
        labels.unsqueeze(dim=-1),
    )


def test_mult_vertices() -> None:
    """Test union find on a 3 and 4 track jet.

    The former with one three track vertex and the latter with two two track vertices.
    """
    labels_jet1 = torch.tensor([0, 0, 2])
    labels_jet2 = torch.tensor([0, 1, 1, 0])
    labels = torch.cat([labels_jet1.flatten(), labels_jet2.flatten()], dim=0)
    pairwise_probs_jet1 = torch.tensor([[1.0, -1.0], [1.0, -1.0], [-1.0, -1.0]])
    pairwise_probs_jet2 = torch.tensor([
        [-1.0, -1.0, 1.0],
        [-1.0, 1.0, -1.0],
        [-1.0, 1.0, -1.0],
        [1.0, -1.0, -1.0],
    ])
    pairwise_probs = torch.cat(
        [pairwise_probs_jet1.flatten(), pairwise_probs_jet2.flatten()], dim=0
    )
    mask = torch.tensor([[False, False, False, True], [False, False, False, False]])
    assert torch.equal(
        get_node_assignment(pairwise_probs.flatten().unsqueeze(dim=-1), mask),
        labels.flatten().unsqueeze(dim=-1),
    )



import torch
from ftag import get_mock_file

from salt.data import SaltDataset


def test_salt_dataset():
    f = get_mock_file()[0]
    norm_dict = {}
    variables = {"jets": ["pt", "eta"], "tracks": ["d0"]}
    dataset = SaltDataset(f, norm_dict, variables, "train")
    for i in range(0, len(dataset), 10):
        dataset[i : i + 10]


def test_track_selector():
    f = get_mock_file()[0]
    norm_dict = {}
    variables = {"jets": ["pt", "eta"], "tracks": ["d0"]}
    selections = {"tracks": ["d0 < 3.5"]}
    no_sel = SaltDataset(f, norm_dict, variables, "train")
    sel = SaltDataset(f, norm_dict, variables, "train", selections=selections)
    for i in range(0, len(no_sel), 10):
        inputs, masks, _ = no_sel[i : i + 10]
        assert torch.any(inputs["tracks"][~masks["tracks"]] > 3.5)
        inputs, masks, _ = sel[i : i + 10]
        assert not torch.any(inputs["tracks"][~masks["tracks"]] > 3.5)



import numpy as np
import torch
from torch.utils.data import Sampler


class RandomBatchSampler(Sampler):
    def __init__(
        self,
        dataset: torch.utils.data.Dataset,
        batch_size: int,
        shuffle: bool = False,
        drop_last: bool = False,
    ):
        """Batch sampler for an h5 dataset.

        The batch sampler performs weak shuffling. Objects are batched first,
        and then batches are shuffled.

        Parameters
        ----------
        dataset : torch.data.Dataset
            Input dataset
        batch_size : int
            Number of objects to batch
        shuffle : bool
            Shuffle the batches
        drop_last : bool
            Drop the last incomplete batch (if present)
        """
        self.batch_size = batch_size
        self.dataset_length = len(dataset)
        self.n_batches = self.dataset_length / self.batch_size
        self.nonzero_last_batch = int(self.n_batches) < self.n_batches
        self.drop_last = drop_last
        self.shuffle = shuffle

    def __len__(self):
        return int(self.n_batches) + int(not self.drop_last and self.nonzero_last_batch)

    def __iter__(self):
        if self.shuffle:
            self.batch_ids = torch.randperm(int(self.n_batches))
        else:
            self.batch_ids = torch.arange(int(self.n_batches))
        # yield full batches from the dataset
        for batch_id in self.batch_ids:
            start, stop = batch_id * self.batch_size, (batch_id + 1) * self.batch_size
            yield np.s_[int(start) : int(stop)]

        # in case the batch size is not a perfect multiple of the number of samples,
        # yield the remaining samples
        if not self.drop_last and self.nonzero_last_batch:
            start, stop = int(self.n_batches) * self.batch_size, self.dataset_length
            yield np.s_[int(start) : int(stop)]



from collections.abc import Mapping

import torch
from torch import Tensor, nn

from salt.models import MaskFormerLoss
from salt.models.transformer_v2 import GLU, Attention
from salt.stypes import Tensors
from salt.utils.mask_utils import indices_from_mask


class MaskDecoder(nn.Module):
    """Mask decoder for Salt - Uses constituent embeddings to generate object queries,
    from which a classification task and mask prediction task, as well as any further object
    level tasks, are performed.

    Based on
    - https://github.com/facebookresearch/MaskFormer
    - https://github.com/facebookresearch/Mask2Former
    """

    def __init__(
        self,
        embed_dim: int,
        num_layers: int,
        md_config: Mapping,
        class_net: nn.Module,
        mask_net: nn.Module,
        num_objects: int,
        loss_config: Mapping,
        aux_loss: bool = False,
    ):
        super().__init__()
        self.aux_loss = aux_loss

        self.inital_q = nn.Parameter(torch.empty((num_objects, embed_dim)))
        nn.init.normal_(self.inital_q)

        self.class_net = class_net
        self.mask_net = mask_net

        self.layers = nn.ModuleList([
            MaskDecoderLayer(embed_dim, mask_net=mask_net, **md_config) for _ in range(num_layers)
        ])
        # Two norm layers may be overkill but it should help with model stability
        self.norm1 = nn.LayerNorm(embed_dim)
        self.norm2 = nn.LayerNorm(embed_dim)

        self.mask_loss = MaskFormerLoss(**loss_config, num_objects=num_objects)

    def get_preds(self, queries: Tensor, mask_tokens: Tensor, pad_mask: Tensor | None = None):
        # get class predictions from queries
        class_logits = self.class_net(queries)
        if class_logits.shape[-1] == 1:
            class_probs = class_logits.sigmoid()
            class_probs = torch.cat([1 - class_probs, class_probs], dim=-1)
        else:
            class_probs = class_logits.softmax(-1)

        # get mask predictions from queries and mask tokens
        pred_masks = get_masks(mask_tokens, queries, self.mask_net, pad_mask)

        return {"class_logits": class_logits, "class_probs": class_probs, "masks": pred_masks}

    def forward(
        self,
        preds: Tensors,
        tasks: nn.ModuleList,
        pad_mask: Tensor = None,
        labels: Tensors | None = None,
    ):
        """Forward pass through the MaskDecoder. Utilises the encoder embeddings to
        generate M query vectors, from which a classification task and mask prediction are
        performed. If labels are provided, the HungarianMatcher algorithm is implemented
        to find the optimal match between predictions and labels. The model predictions are
        then re-ordered to match the labels, and a final loss is calculated.

        Parameters
        ----------
        preds : dict[str, Tensor]
            Dictionary containing existing predictions from the encoder, should contain the
            key 'embed_xs' which is the encoder embeddings.
        tasks : nn.ModuleList
            The tasks for the model to perform, excluding the object mask prediction and
            classification. Any tasks that use the input stream 'objects' will be used
        pad_mask : Tensor, optional
            Masks for input padding, by default None
        labels : dict[str, Tensor] | None, optional
            Dict containing labels, by default None

        Returns
        -------
        preds : dict[str, Tensor]
            Dictionary containing the updated model predictions, with all object predictions
            stored in preds["objects"]
        labels : dict[str, Tensor] | None
            Updated labels, containing all default labels and any additional labels generated
            by the model (e.g. regression targets). If input labels are None, returns None.
        loss :
            Loss value if labels are provided, otherwise returns an empty dict.
        """
        # MF only supports one input, if we have multiple then we have no way of knowing
        # what section of the embedding relates to objects we want to generate masks for
        if isinstance(pad_mask, dict):
            assert len(pad_mask) == 1, "Maskformer only supports one input."
            pad_mask = next(iter(pad_mask.values()))

        x = preds["embed_xs"]
        # apply norm
        q = self.norm1(self.inital_q.expand(x.shape[0], -1, -1))
        x = self.norm2(x)

        # Add a dummy track to the inputs (and to pad mask) to stop onnx complaining
        xpad = torch.zeros((x.shape[0], 1, x.shape[-1]), device=x.device, dtype=x.dtype)
        x = torch.cat([x, xpad], dim=1)
        if pad_mask is not None:
            padpad_mask = torch.zeros(
                (pad_mask.shape[0], 1), device=pad_mask.device, dtype=pad_mask.dtype
            )
            pad_mask = torch.cat([pad_mask, padpad_mask], dim=1)

        intermediate_outputs: list | None = [] if self.aux_loss else None
        for layer in self.layers:
            if self.aux_loss:
                assert intermediate_outputs is not None
                intermediate_outputs.append({"embed": q, **self.get_preds(q, x, pad_mask)})
            q, x = layer(q, x, kv_mask=pad_mask)
        mf_preds = self.get_preds(q, x, pad_mask)

        # Un-pad the embedding x, get the mf_predictions, and then unpad them as well
        preds["objects"] = {"embed": q, "x": x[:, :-1, :], **mf_preds}
        preds["objects"]["masks"] = preds["objects"]["masks"][:, :, :-1]
        if self.aux_loss:
            preds["intermediate_outputs"] = intermediate_outputs

        if labels is not None:
            return self.mask_loss(preds, tasks, labels)

        return preds, labels, {}


def get_masks(x: Tensor, q: Tensor, mask_net: nn.Module, input_pad_mask: Tensor | None = None):
    mask_tokens = mask_net(q)
    pred_masks = torch.einsum("bqe,ble->bql", mask_tokens, x)

    if input_pad_mask is not None:
        pred_masks[input_pad_mask.unsqueeze(1).expand_as(pred_masks)] = torch.finfo(
            pred_masks.dtype
        ).min

    return pred_masks


def get_maskformer_outputs(
    objects,
    max_null=0.5,
    apply_reorder=True,
):
    """Takes objects for a single MF global prediction and converts them to a more useful
    form.

        -  Keep only objects which have `pnull < max_null`
        - Converts the mask logits to mask indices, see salt.utils.mask_utils.indices_from_mask

    """
    # Convert the (N,M) -> (M,) mask indices
    masks = objects["masks"]
    class_probs = objects["class_probs"]
    regression = objects["regression"]
    n_tracks = masks.shape[-1]
    n_obj = masks.shape[1]
    n_reg = regression.shape[-1]

    # If we have a jet with no tracks,
    if n_tracks == 0:
        return (
            torch.full((1, n_obj), torch.nan),
            None,
            class_probs,
            torch.full((1, n_obj, n_reg), torch.nan),
        )
    # For testing purposes - this will likely blow up our fake rate
    null_preds = class_probs[:, :, -1] > max_null
    if not null_preds.any():
        # If we have no predicted objects, we return dummy values
        return (
            torch.full((1, n_obj), torch.nan),
            torch.arange(n_tracks).unsqueeze(0).expand(1, n_tracks),
            class_probs,
            torch.full((1, n_obj, n_reg), torch.nan),
        )

    masks = masks.sigmoid() > 0.5
    expanded_null = null_preds.unsqueeze(-1).expand(-1, -1, masks.size(-1))
    masks[expanded_null] = torch.zeros_like(masks)[expanded_null]
    regression[null_preds] = torch.nan

    if apply_reorder:
        # Define the leading object as the one with the highest regression[0] value
        # in vertexing case, this is the pT. We first set these values to non-nans, as
        # argsort will otherwise not work correctly when in athena, and then set them back
        regression[null_preds] = -torch.inf
        order = torch.argsort(regression[:, :, 0], descending=True)
        regression[null_preds] = torch.nan
        order_expanded = order.unsqueeze(-1).expand(-1, -1, masks.size(-1))

        # Use gather to reorder tensors along a specific dimension
        masks = torch.gather(masks, 1, order_expanded)
        class_probs = torch.gather(
            class_probs, 1, order.unsqueeze(-1).expand(-1, -1, class_probs.size(-1))
        )
        regression = torch.gather(
            regression, 1, order.unsqueeze(-1).expand(-1, -1, regression.size(-1))
        )
        # Define the leading object as that with the highest [0] (pt for vertexing)
    leading_regression = regression[:, 0]

    # Convert our masks (N,M), now in pT order, to be (M,) indices
    obj_indices = indices_from_mask(masks)

    return leading_regression, obj_indices, class_probs, regression


class MaskDecoderLayer(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        n_heads: int,
        mask_attention: bool,
        bidirectional_ca: bool,
        mask_net: nn.Module,
    ) -> None:
        super().__init__()

        self.mask_attention = mask_attention
        self.bidirectional_ca = bidirectional_ca

        self.q_ca = Attention(embed_dim=embed_dim, num_heads=n_heads)
        self.q_sa = Attention(embed_dim=embed_dim, num_heads=n_heads)
        self.q_dense = GLU(embed_dim)
        if bidirectional_ca:
            self.kv_ca = Attention(embed_dim=embed_dim, num_heads=n_heads)
            self.kv_dense = GLU(embed_dim)
        self.mask_net = mask_net

    def forward(self, q: Tensor, kv: Tensor, kv_mask: Tensor | None = None) -> Tensor:
        attn_mask = None
        # return q, kv
        # if we want to do mask attention
        if self.mask_attention:
            # New attention masking convention with transformers 2
            # Positions with True are allowed while False are masked
            # Compute masks and apply sigmoid
            attn_mask = get_masks(kv, q, self.mask_net, kv_mask).sigmoid()

            # Threshold and detach
            attn_mask = (attn_mask > 0.9).detach()
            # Check if all values along the last dimension are 0 (equivalent to `False` in boolean)
            # If so, set them to 1 (equivalent to `True` in boolean)
            newmask = torch.all(attn_mask == 0, dim=-1, keepdim=True).expand(attn_mask.shape)

            attn_mask = attn_mask | newmask

        # update queries with cross attention from nodes
        q = q + self.q_ca(q, kv=kv, kv_mask=kv_mask, attn_mask=attn_mask)

        # update queries with self attention
        q = q + self.q_sa(q)

        # dense update
        q = q + self.q_dense(q)

        # update nodes with cross attention from queries and dense layer
        if self.bidirectional_ca:
            if attn_mask is not None:
                attn_mask = attn_mask.transpose(1, 2)
                newmask = torch.all(attn_mask == 1, dim=-1, keepdim=True).expand(attn_mask.shape)
                attn_mask = attn_mask | ~newmask.bool()

            kv = kv + self.kv_ca(kv, q, attn_mask=attn_mask)
            kv = kv + self.kv_dense(kv)
        return q, kv



from salt.models.attention import GATv2Attention, MultiheadAttention, ScaledDotProductAttention
from salt.models.dense import Dense
from salt.models.featurewise import FeaturewiseTransformation
from salt.models.initnet import InitNet
from salt.models.inputnorm import InputNorm
from salt.models.maskformer_loss import MaskFormerLoss
from salt.models.pooling import (
    DictCrossAttentionPooling,
    GlobalAttentionPooling,
    NodeQueryGAP,
    Pooling,
    TensorCrossAttentionPooling,
)
from salt.models.posenc import PositionalEncoder
from salt.models.r21xbb import R21Xbb
from salt.models.saltmodel import SaltModel
from salt.models.task import (
    ClassificationTask,
    GaussianRegressionTask,
    RegressionTask,
    RegressionTaskBase,
    TaskBase,
    VertexingTask,
)
from salt.models.transformer import (
    TransformerCrossAttentionEncoder,
    TransformerCrossAttentionLayer,
    TransformerEncoder,
    TransformerEncoderLayer,
)
from salt.models.transformer_v2 import TransformerV2

# from salt.models.maskformer import MaskFormer

__all__ = [
    "ClassificationTask",
    "Dense",
    "DictCrossAttentionPooling",
    "FeaturewiseTransformation",
    "GATv2Attention",
    "GaussianRegressionTask",
    "GlobalAttentionPooling",
    "InitNet",
    "InputNorm",
    "MaskFormer",
    "MaskFormerLoss",
    "MultiheadAttention",
    "NodeQueryGAP",
    "Pooling",
    "PositionalEncoder",
    "R21Xbb",
    "RegressionTask",
    "RegressionTaskBase",
    "SaltModel",
    "ScaledDotProductAttention",
    "TaskBase",
    "TensorCrossAttentionPooling",
    "Transformer",
    "TransformerCrossAttentionEncoder",
    "TransformerCrossAttentionLayer",
    "TransformerEncoder",
    "TransformerEncoderLayer",
    "TransformerV2",
    "VertexingTask",
]



import logging
import subprocess
from pathlib import Path
from typing import Any

logging.basicConfig(level=logging.INFO)


class SlurmHandler:
    """A class to submit batch jobs to a Slurm scheduler.

    Attributes
    ----------
    batch_path : Path
        Path where the batch file which is created will be stored.
    log_path : Path
        Path where the batch log files will be stored.
    base_dir : Path
        Directory in which batch job will execute its command.

    Methods
    -------
    activate_testmode():
        Activate test mode: check config files in dry runs, no jobs submitted.
    deactivate_testmode():
        Deactivate test mode, enable submitting jobs.
    send_job(command: str, tag: str = "slurm_job"):
        Submit job by creating and executing Slurm batch file
    """

    def __init__(self, batch_path: str, log_path: str, basedir: str) -> None:
        self.batch_path = Path(batch_path)
        self.log_path = Path(log_path)
        self.base_dir = Path(basedir) if basedir else Path.cwd()
        self._tag = "salt_job"
        # Keywords to be used in Slurm configuration
        self._slurm_options_dict: dict[str, Any] = {}
        self._test_mode = False

    def activate_testmode(self) -> None:
        logging.debug("Activated test mode: not submitting any jobs.")
        self._test_mode = True

    def deactivate_testmode(self) -> None:
        logging.debug("Deactivated test mode: submitting jobs.")
        self._test_mode = False

    def send_job(self, command: str, tag: str = "salt_job") -> None:
        self._tag = tag
        batchfile = self._make_batch_file(command)
        if self._test_mode:
            logging.debug(f"Created batch file {batchfile}")
        else:
            subprocess.call(f"sbatch {batchfile}", shell=True)

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        self._slurm_options_dict[key] = value

    def _make_batch_file(self, command: str) -> Path:
        batch_file = self.batch_path / f"sbatch_{self._tag}.sh"
        with batch_file.open("w") as bf:
            bf.write(f"""#!/bin/sh
# {self._tag} batch run script\n""")
            for key, value in self._slurm_options_dict.items():
                if value is None:
                    bf.write(f"#SBATCH --{key}\n")
                else:
                    bf.write(f"#SBATCH --{key}={value}\n")
            bf.write(f"""BASEDIR={self.base_dir};pwd; ls -l\n""")
            bf.write(f"""{command}""")
        batch_file.chmod(0o755)
        logging.debug(f"Made batch file {batch_file}")
        return batch_file



"""Adapted from the muP GitHub Repo
[1] https://github.com/microsoft/mup
Helper functions for performing coord check.
"""

from copy import copy
from pathlib import Path
from time import time

import pandas as pd
import torch

FDICT = {
    "l1": lambda x: torch.abs(x).mean(),
    "l2": lambda x: (x**2).mean() ** 0.5,
    "mean": lambda x: x.mean(),
    "std": lambda x: x.std(),
}


def convert_fdict(d):
    """Convert a dict `d` with string values to function values.

    Input:
        d: a dict whose values are either strings or functions
    Output:
        a new dict, with the same keys as `d`, but the string values are
        converted to functions using `FDICT`.
    Source: [1]
    """
    return dict([((k, FDICT[v]) if isinstance(v, str) else (k, v)) for k, v in d.items()])


def fix_batch_dataloader(batch, nsteps):
    """Create a fake dataloader with nsteps times the same batch.
    These nsteps versions are indepedent of each others for the
    inputs.

    This is required as our batch are list of dictionaries, so
    we need to deepcopy the batch to avoid having a pass of the model
    changing all batches in the fake dataloader.

    Note: copy.deecopy is not enough!
    """
    dataloader = []
    for _ in range(nsteps):
        new_batch = []
        for count, entry in enumerate(batch):
            new_entry = {}
            for key, val in entry.items():
                if count != 0:
                    new_entry[key] = val
                else:
                    new_entry[key] = val.clone().detach()
            new_batch.append(new_entry)
        dataloader.append(new_batch)
    return dataloader


def _record_coords(
    records, width, modulename, t, output_fdict=None, input_fdict=None, param_fdict=None
):
    """Returns a forward hook that records coordinate statistics.
    Adapted from muP GitHub Repo [1].

    Returns a forward hook that records statistics regarding the output, input,
    and/or parameters of a `nn.Module`. This hook is intended to run only once,
    on the timestep specified by `t`.

    On forward pass, the returned hook calculates statistics specified in
    `output_fdict`, `input_fdict`, and `param_fdict`, such as the normalized l1
    norm, of output, input, and/or parameters of the module. The statistics are
    recorded along with the `width`, `modulename`, and `t` (the time step) as a
    dict and inserted into `records` (which should be a list). More precisely,
    for each output, input, and/or parameter, the inserted dict is of the form

        {
            'width': width, 'module': modified_modulename, 't': t,
            # keys are keys in fdict
            'l1': 0.241, 'l2': 0.420, 'mean': 0.0, ...
        }

    where `modified_modulename` is a string that combines the `modulename` with
    an indicator of which output, input, or parameter tensor is the statistics
    computed over.

    The `*_fdict` inputs should be dictionaries with string keys and whose
    values can either be functions or strings. The string values are converted
    to functions via `convert_fdict`. The default values of `*_dict` inputs are
    converted to `output_fdict = dict(l1=FDICT['l1'])`, `input_fdict = {}`,
    `param_fdict = {}`, i.e., only the average coordinate size (`l1`) of the
    output activations are recorded.

    Inputs:
        records:
            list to append coordinate data to
        width:
            width of the model. This is used only for plotting coord check later
            on, so it can be any notion of width.
        modulename:
            string name of the module. This is used only for plotting coord check.
        t:
            timestep of training. This is used only for plotting coord check.
        output_fdict, input_fdict, param_fdict:
            dicts with string keys and whose values can either be functions or
            strings. The string values are converted to functions via
            `convert_fdict`
    Output:
        a forward hook that records statistics regarding the output, input,
        and/or parameters of a `nn.Module`, as discussed above.
    """
    output_fdict = {"l1": FDICT["l1"]} if output_fdict is None else convert_fdict(output_fdict)
    input_fdict = {} if input_fdict is None else convert_fdict(input_fdict)
    param_fdict = {} if param_fdict is None else convert_fdict(param_fdict)

    def f(module, myinput, myoutput):
        def get_stat(d, x, fdict):
            if isinstance(x, tuple | list):
                for i, _x in enumerate(x):
                    _d = copy(d)
                    _d["module"] += f"[{i}]"
                    get_stat(_d, _x, fdict)
            elif isinstance(x, dict):
                for name, _x in x.items():
                    _d = copy(d)
                    _d["module"] += f"[{name}]"
                    get_stat(_d, _x, fdict)
            elif isinstance(x, torch.Tensor):
                _d = copy(d)
                for fname, f in fdict.items():
                    _d[fname] = f(x).item()
                records.append(_d)
            else:
                raise NotImplementedError(f"Unexpected output type: {type(x)}")

        with torch.no_grad():
            ret = {"width": width, "module": modulename, "t": t}

            # output stats
            if isinstance(myoutput, tuple | list):
                for i, out in enumerate(myoutput):
                    _ret = copy(ret)
                    _ret["module"] += f":out[{i}]"
                    get_stat(_ret, out, output_fdict)
            elif isinstance(myoutput, dict):
                for name, out in myoutput.items():
                    _ret = copy(ret)
                    _ret["module"] += f":out[{name}]"
                    get_stat(_ret, out, output_fdict)
            elif isinstance(myoutput, torch.Tensor):
                _ret = copy(ret)
                for fname, f in output_fdict.items():
                    _ret[fname] = f(myoutput).item()
                records.append(_ret)
            else:
                raise NotImplementedError(f"Unexpected output type: {type(myoutput)}")

            # input stats
            if input_fdict:
                if isinstance(myinput, tuple | list):
                    for i, out in enumerate(myinput):
                        _ret = copy(ret)
                        _ret["module"] += f":in[{i}]"
                        get_stat(_ret, out, input_fdict)
                elif isinstance(myinput, dict):
                    for name, out in myinput.items():
                        _ret = copy(ret)
                        _ret["module"] += f":in[{name}]"
                        get_stat(_ret, out, input_fdict)
                elif isinstance(myinput, torch.Tensor):
                    _ret = copy(ret)
                    for fname, f in input_fdict.items():
                        _ret[fname] = f(myinput).item()
                    records.append(_ret)
                else:
                    raise NotImplementedError(f"Unexpected output type: {type(myinput)}")

            # param stats
            if param_fdict:
                for name, p in module.named_parameters():
                    _ret = copy(ret)
                    _ret["module"] += f":param[{name}]"
                    for fname, f in param_fdict.items():
                        _ret[fname] = f(p).item()
                    records.append(_ret)

    return f


def _get_coord_data(
    mup,
    models,
    dataloader,
    optcls,
    nsteps=3,
    filter_module_by_name=None,
    fix_data=True,
    cuda=False,
    nseeds=1,
    output_fdict=None,
    input_fdict=None,
    param_fdict=None,
    show_progress=True,
):
    """Inner method for `get_coord_data`, adapted from [1].

    Train the models in `models` with optimizer given by `optcls` and data from
    `dataloader` for `nsteps` steps, and record coordinate statistics specified
    by `output_fdict`, `input_fdict`, `param_fdict`. By default, only `l1` is
    computed for output activations of each module.

    Inputs:
        mup: bool,
            whether mup is to be used. If so, set_base_shape with rescale_params=rescale
        models:
            a dict of model paths, where the keys are numbers indicating width.
            Each entry of `models` is a path to an already instantiated model.
        dataloader:
            a dataloader, compatible with the global object classification task.
        optcls:
            a function so that `optcls(model)` gives an optimizer used to train
            the model.
        nsteps:
            number of steps to train the model
        filter_module_by_name:
            a function that returns a bool given module names (from
            `model.named_modules()`), or None. If not None, then only modules
            whose name yields True will be recorded.
        fix_data:
            Whether to fix the dataloader to a single batch. Default: True
        cuda:
            whether to use cuda or not. Default: False
        nseeds:
            number of times to repeat the training, each with different seeds.
        output_fdict, input_fdict, param_fdict:
            function dicts to be used in `_record_coords`. By default, only `l1`
            is computed for output activations of each module.
        show_progress:
            show progress using tqdm.
    Output:
        a pandas DataFrame containing recorded results. The column names are
        `'width', 'module', 't'` as well as names of statistics recorded, such
        as `'l1'` (see `FDICT` for other premade statistics that can be
        collected).
    """
    df = []
    if fix_data:
        saved_batch = next(iter(dataloader))

    if show_progress:
        from tqdm import tqdm

        pbar = tqdm(total=nseeds * len(models))

    for i in range(nseeds):
        torch.manual_seed(i)
        for width, model_path in models.items():
            if fix_data:
                dataloader = fix_batch_dataloader(saved_batch, nsteps)
            model = torch.load(model_path)
            model.train()
            if mup:
                from salt.utils.muP_utils.configuration_muP import instantiate_mup

                instantiate_mup(model, check=True)

            if cuda:
                model.cuda()
            optimizer = optcls(model.model)
            for batch_idx, batch in enumerate(dataloader, 1):
                remove_hooks = []
                # add hooks
                for name, module in model.named_modules():
                    if filter_module_by_name and not filter_module_by_name(name):
                        continue
                    remove_hooks.append(
                        module.register_forward_hook(
                            _record_coords(
                                df,
                                width,
                                name,
                                batch_idx,
                                output_fdict=output_fdict,
                                input_fdict=input_fdict,
                                param_fdict=param_fdict,
                            )
                        )
                    )

                if cuda:
                    batch.cuda()

                _, _, _, loss = model.shared_step(batch)
                optimizer.zero_grad()
                loss["loss"].backward()
                optimizer.step()

                # remove hooks
                for handle in remove_hooks:
                    handle.remove()

                if batch_idx == nsteps:
                    break
            if show_progress:
                pbar.update(1)
    if show_progress:
        pbar.close()
    return pd.DataFrame(df)


def get_coord_data(models, dataloader, optimizer="adamw", lr=None, mup=True, **kwargs):
    """Get coord data for coord check.

    Train the models in `models` with data from `dataloader` and optimizer
    specified by `optimizer` and `lr` for `nsteps` steps, and record coordinate
    statistics specified by `output_fdict`, `input_fdict`, `param_fdict`. By
    default, only `l1` is computed for output activations of each module.

    This function wraps around `_get_coord_data`, with the main difference being
    user can specify common optimizers via a more convenient interface.

    Inputs:
        models:
            a dict of lazy models, where the keys are numbers indicating width.
            Each entry of `models` is a path to an already instantiated model.
        dataloader:
            a dataloader, compatible with the global object classification task.
        optimizer:
            a string in `['sgd', 'adam', 'adamw']`, with default being `'adamw'`.
        lr:
            learning rate. By default is 0.1 for `'sgd'` and 1e-2 for others.
        mup:
            If True, then use the optimizer from `mup.optim`; otherwise, use the
            one from `torch.optim`.
        filter_trainable_by_name:
            a function that returns a bool given module names (from
            `model.named_modules()`), or None. If not None, then only modules
            whose name yields True will be trained.
        nsteps:
            number of steps to train the model
        filter_module_by_name:
            a function that returns a bool given module names (from
            `model.named_modules()`), or None. If not None, then only modules
            whose name yields True will be recorded.
        fix_data:
            Whether to fix the dataloader to a single batch. Default: True
        cuda:
            whether to use cuda or not. Default: True
        nseeds:
            number of times to repeat the training, each with different seeds.
        output_fdict, input_fdict, param_fdict:
            function dicts to be used in `_record_coords`. By default, only `l1`
            is computed for output activations of each module.
        show_progress:
            show progress using tqdm.
    Output:
        a pandas DataFrame containing recorded results. The column names are
        `'width', 'module', 't'` as well as names of statistics recorded, such
        as `'l1'` (see `FDICT` for other premade statistics that can be
        collected).
    """
    if lr is None:
        lr = 0.1 if optimizer == "sgd" else 1e-2
    if mup:
        from mup.optim import MuAdam as Adam
        from mup.optim import MuAdamW as AdamW
        from mup.optim import MuSGD as SGD  # noqa: N814
    else:
        from torch.optim import SGD, Adam, AdamW

    def get_trainable(model):
        params = []
        for _, p in model.named_parameters():
            params.append(p)
        return params

    if optimizer == "sgd":

        def optcls(model):
            return SGD(get_trainable(model), lr=lr)

    elif optimizer == "adam":

        def optcls(model):
            return Adam(get_trainable(model), lr=lr)

    elif optimizer == "adamw":

        def optcls(model):
            return AdamW(get_trainable(model), lr=lr, weight_decay=1e-5)

    data = _get_coord_data(mup, models, dataloader, optcls, **kwargs)
    data["optimizer"] = optimizer
    data["lr"] = lr
    return data


def plot_coord_data(
    df,
    y="l1",
    save_to=None,
    suptitle=None,
    x="width",
    hue="module",
    legend="full",
    name_contains=None,
    name_not_contains=None,
    loglog=True,
    logbase=2,
    face_color=None,
):
    """Plot coord check data `df` obtained from `get_coord_data`.

    Input:
        df:
            a pandas DataFrame obtained from `get_coord_data`
        y:
            the column of `df` to plot on the y-axis. Default: `'l1'`
        save_to:
            path to save the resulting figure, or None. Default: None.
        suptitle:
            The title of the entire figure.
        x:
            the column of `df` to plot on the x-axis. Default: `'width'`
        hue:
            the column of `df` to represent as color. Default: `'module'`
        legend:
            'auto', 'brief', 'full', or False. This is passed to `seaborn.lineplot`.
        name_contains:
            only plot modules whose name contains `name_contains`
        name_not_contains:
            only plot modules whose name does not contain `name_not_contains`
        loglog:
            whether to use loglog scale. Default: True
        logbase:
            the log base, if using loglog scale. Default: 2
        face_color:
            background color of the plot. Default: None (which means white)
    Output:
        the `matplotlib` figure object
    """
    # preprocessing
    df = copy(df)
    # nn.Sequential has name '', which duplicates the output layer
    df = df[df.module != ""]
    try:
        if name_contains is not None:
            if not isinstance(name_contains, list):
                name_contains = [name_contains]
            df = df[df["module"].str.contains("|".join(name_contains))]
        if name_not_contains is not None:
            if not isinstance(name_not_contains, list):
                name_not_contains = [name_not_contains]
            df = df[~(df["module"].str.contains("|".join(name_not_contains)))]
        # for nn.Sequential, module names are numerical
        df["module"] = pd.to_numeric(df["module"])
    except Exception:  # noqa: S110, BLE001
        pass

    ts = df.t.unique()

    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set()

    def tight_layout(plt):
        plt.tight_layout(rect=[0, 0.03, 1, 0.95])

    # plot
    fig = plt.figure(figsize=(5 * len(ts), 4))
    if face_color is not None:
        fig.patch.set_facecolor(face_color)
    for t in ts:
        plt.subplot(1, len(ts), t)
        sns.lineplot(x=x, y=y, data=df[df.t == t], hue=hue, legend=legend if t == 1 else None)
        plt.title(f"t={t}")
        if t != 1:
            plt.ylabel("")
        if loglog:
            plt.loglog(base=logbase)
    if suptitle:
        suptitle = (
            f"{suptitle} for layers {' & '.join(name_contains)}"
            if name_contains is not None
            else suptitle
        )
        suptitle = (
            f"{suptitle} without layers {' & '.join(name_not_contains)}"
            if name_not_contains is not None
            else suptitle
        )
        plt.suptitle(suptitle)
    tight_layout(plt)
    if save_to is not None:
        plt.savefig(save_to)
        print(f"Coordinate check plot saved to {save_to}")

    return fig


def _get_training_data(
    mup,
    models,
    dataloader,
    optcls,
    save_model=None,
    nsteps=10,
    cuda=True,
    nseeds=1,
    show_progress=True,
    noTQDM=False,
):
    """Inner method for `get_training_data`.

    Train the models in `models` with optimizer given by `optcls` and data from
    `dataloader` for `nsteps` steps, and record loss.

    Inputs:
        mup: bool,
            Whether mup is to be used. If so, set_base_shape with rescale_params=rescale
        models:
            a dict of model paths, where the keys are numbers indicating width.
            Each entry of `models` is a path to an already instantiated model.
        dataloader:
            a dataloader, compatible with the global object classification task.
        optcls:
            a function so that `optcls(model)` gives an optimizer used to train
            the model.
        save_model:
            path to save the model (if not None)
        nsteps:
            number of steps to train the model
        cuda:
            whether to try and use cuda. Default: True
        nseeds:
            number of times to repeat the training, each with different seeds.
        show_progress:
            show progress (either printed or TQDM, depending on next param)
        noTQDM:
            do not show progress using tqdm. Default: True
    Output:
        a pandas DataFrame containing recorded results. The column names are
        `'width', 'step', 'loss'`.
    """
    df = pd.DataFrame(columns=["width", "step", "loss"])

    def move_tensors_to_gpu(data):
        if isinstance(data, dict):
            return {key: move_tensors_to_gpu(value) for key, value in data.items()}
        if isinstance(data, list):
            return [move_tensors_to_gpu(item) for item in data]
        if isinstance(data, torch.Tensor):
            return data.cuda()
        return data

    if torch.cuda.is_available() and cuda:
        print("Using GPU")
        cuda = True
    else:
        print("Using CPU")
        cuda = False

    if save_model is not None:
        save_path = Path(save_model)
        save_path.mkdir(exist_ok=True)
        freq = 1 if nsteps <= 5 else round(nsteps / 5) if nsteps <= 1000 else round(nsteps / 10)

    freq_bar = 1 if nsteps <= 5 else round(nsteps / 10) if nsteps <= 1000 else round(nsteps / 100)
    for i in range(nseeds):
        torch.manual_seed(i)
        for width, model_path in models.items():
            if show_progress and not noTQDM:
                from tqdm import tqdm

                pbar = tqdm(total=nsteps)

            model = torch.load(model_path)
            model.train()

            if mup:
                from salt.utils.muP_utils.configuration_muP import instantiate_mup

                instantiate_mup(model, check=True)

            if cuda:
                model.cuda()
            optimizer = optcls(model.model)

            for batch_idx, batch in enumerate(dataloader, 1):
                t0 = time()
                if batch_idx < 10:
                    tf0 = time()

                mybatch = move_tensors_to_gpu(batch) if cuda else batch
                _, _, _, loss = model.shared_step(mybatch)
                optimizer.zero_grad()
                loss["loss"].backward()
                optimizer.step()

                myloss = loss["loss"].detach().cpu().numpy()
                df.loc[len(df)] = [width, batch_idx, myloss]

                if (save_model is not None) and (batch_idx % freq == 0):
                    torch.save(model, save_path / f"width_{width}_seed_{i}_step_{batch_idx}.pt")

                if show_progress and (batch_idx % freq_bar == 0):
                    t1 = time()
                    if not noTQDM:
                        pbar.update(freq_bar)
                    else:
                        print(
                            f"Step {batch_idx} / {nsteps} - loss = {myloss:.5} in time {t1 - t0:.3}"
                        )
                    t0 = t1
                if batch_idx < 10:
                    tf1 = time()
                    print(f"Iteration number {batch_idx} done in {tf1 - tf0:.3}")

                if batch_idx == nsteps:
                    break

            if save_model is not None:
                torch.save(model, save_path / f"width_{width}_seed_{i}_step_{batch_idx}.pt")
    if show_progress and not noTQDM:
        pbar.close()
    return df


def get_training_data(models, dataloader, optimizer="adamw", lr=None, mup=True, **kwargs):
    """Get training data.

    Train the models in `models` with data from `dataloader` and optimizer
    specified by `optimizer` and `lr` for `nsteps` steps, and record loss.

    This function wraps around `_get_coord_data`, with the main difference being
    user can specify common optimizers via a more convenient interface.

    Inputs:
        models:
            a dict of model paths, where the keys are numbers indicating width.
            Each entry of `models` is a path to an already instantiated model.
        dataloader:
            a dataloader, compatible with the jet classification task.
        optimizer:
            a string in `['sgd', 'adam', 'adamw']`, with default being `'adamw'`.
        lr:
            learning rate. By default is 0.1 for `'sgd'` and 1e-2 for others.
        mup:
            If True, then use the optimizer from `mup.optim`; otherwise, use the
            one from `torch.optim`.
        nsteps:
            number of steps to train the model
        cuda:
            whether to use cuda or not. Default: True
        nseeds:
            number of times to repeat the training, each with different seeds.
        show_progress:
            show progress using tqdm.
    Output:
        a pandas DataFrame containing recorded results. The column names are
        `'width', 'module', 'loss'`.
    """
    if lr is None:
        lr = 0.1 if optimizer == "sgd" else 1e-2
    if mup:
        from mup.optim import MuAdam as Adam
        from mup.optim import MuAdamW as AdamW
        from mup.optim import MuSGD as SGD  # noqa: N814
    else:
        from torch.optim import SGD, Adam, AdamW

    def get_trainable(model):
        params = []
        for _, p in model.named_parameters():
            params.append(p)
        return params

    if optimizer == "sgd":

        def optcls(model):
            return SGD(get_trainable(model), lr=lr)

    elif optimizer == "adam":

        def optcls(model):
            return Adam(get_trainable(model), lr=lr)

    elif optimizer == "adamw":

        def optcls(model):
            return AdamW(get_trainable(model), lr=lr, weight_decay=1e-5)

    return _get_training_data(mup, models, dataloader, optcls, **kwargs)


def plot_training_data(
    df,
    y="loss",
    save_to=None,
    suptitle=None,
    x="step",
    hue="width",
    legend="full",
    face_color=None,
    window_size=30,
):
    """Plot training check data `df` obtained from `get_training_data`.

    Input:
        df:
            a pandas DataFrame obtained from `get_coord_data`
        y:
            the column of `df` to plot on the y-axis. Default: `'l1'`
        save_to:
            path to save the resulting figure, or None. Default: None.
        suptitle:
            The title of the entire figure.
        x:
            the column of `df` to plot on the x-axis. Default: `'width'`
        hue:
            the column of `df` to represent as color. Default: `'width'`
        legend:
            'auto', 'brief', 'full', or False. This is passed to `seaborn.lineplot`.
        name_contains:
            only plot modules whose name contains `name_contains`
        name_not_contains:
            only plot modules whose name does not contain `name_not_contains`
        loglog:
            whether to use loglog scale. Default: True
        logbase:
            the log base, if using loglog scale. Default: 2
        face_color:
            background color of the plot. Default: None (which means white)
        window_size:
            an int indicating the size of the rolling average of the loss.
    Output:
        the `matplotlib` figure object
    """
    # preprocessing
    import numpy as np

    df = copy(df)

    # Combined the seeds
    df = df.groupby(["width", "step"], sort=False).agg({"loss": np.average})
    df = df.reset_index()

    # Now put a rolling average of loss, based on step for each width
    if window_size is not None:
        df = (
            df.groupby("width")
            .apply(
                lambda x: x.set_index("step")["loss"]
                .rolling(window=window_size, min_periods=1)
                .mean()
                .reset_index(name="loss")
            )
            .reset_index()
            .drop(columns="level_1")
        )

    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set()

    def tight_layout(plt):
        plt.tight_layout(rect=[0, 0.03, 1, 0.95])

    # plot
    fig = plt.figure(figsize=(5, 4))
    if face_color is not None:
        fig.patch.set_facecolor(face_color)
    plt.subplot(1, 1, 1)
    sns.lineplot(x=x, y=y, data=df, hue=hue, legend=legend, palette="flare")
    if suptitle:
        plt.suptitle(suptitle)
    tight_layout(plt)
    if save_to is not None:
        plt.savefig(save_to)
        print(f"Coordinate check plot saved to {save_to}")

    return fig



import torch
from torch import BoolTensor, Tensor


def build_target_masks(object_ids, input_ids, shuffle=False):
    """Get the truth masks from the object ids and input_ids.

    The difference between this function and mask_from_indices is that mask_from_indices
    expects the indices to start from zero, and here we match based on arbitrary IDs,
    such as barcodes.

    Parameters
    ----------
    object_ids : Tensor
        The unqiue ids of the truth object labels
    input_ids : Tensor
        The ids of the per-input labels
    shuffle: bool
        Shuffle object ids

    Returns
    -------
    Tensor
        The truth masks
    """
    # shuffling doesn't seem to be needed here
    if shuffle:
        object_ids = object_ids[:, torch.randperm(object_ids.shape[1])]
    object_ids[object_ids == -1] = -999
    return input_ids.unsqueeze(-2) == object_ids.unsqueeze(-1)


def mask_from_indices(indices: Tensor, num_masks: int | None = None) -> BoolTensor:
    """Convert a dense index tensor to a sparse bool mask.

    Indices are arbitrary and start from 0.

    Examples
    --------
    [0, 1, 1] -> [[True, False, False], [False, True, True]]
    [0, 1, 2]] -> [[True, False, False], [False, True, False], [False, False, True]]

    Parameters
    ----------
    indices : Tensor
        The dense indices
    num_masks : int
        The maximum number of masks

    Returns
    -------
    BoolTensor
        The sparse mask
    """
    assert indices.ndim in {1, 2}, "indices must be 1D for single sample or 2D for batch"
    if num_masks is None:
        num_masks = indices.max() + 1
    else:
        assert (
            num_masks > indices.max()
        ), "num_masks must be greater than the maximum value in indices"

    indices = torch.as_tensor(indices)
    kwargs = {"dtype": torch.bool, "device": indices.device}
    if indices.ndim == 1:
        mask = torch.zeros((num_masks, indices.shape[-1]), **kwargs)
        mask[indices, torch.arange(indices.shape[-1])] = True
        mask.transpose(0, 1)[indices < 0] = False  # handle negative indices
    else:
        mask = torch.zeros((indices.shape[0], num_masks, indices.shape[-1]), **kwargs)
        mask[
            torch.arange(indices.shape[0]).unsqueeze(-1), indices, torch.arange(indices.shape[-1])
        ] = True
        mask.transpose(1, 2)[indices < 0] = False  # handle negative indices

    return mask


def indices_from_mask(mask: BoolTensor, noindex: int = -2) -> Tensor:
    """Convert a sparse bool mask to a dense index tensor.

    Indices are arbitrary and start from 0.

    Examples
    --------
    [[True, False, False], [False, True, True]] -> [0, 1, 1]

    Parameters
    ----------
    mask : BoolTensor
        The sparse mask
    noindex : int
        The value to use for no index

    Returns
    -------
    Tensor
        The dense indices
    """
    mask = torch.as_tensor(mask)
    kwargs = {"dtype": torch.long, "device": mask.device}
    if mask.ndim == 2:
        indices = torch.ones(mask.shape[-1], **kwargs) * noindex
        nonzero_idx = torch.where(mask)
        indices[nonzero_idx[1]] = nonzero_idx[0]
    elif mask.ndim == 3:
        indices = torch.ones((mask.shape[0], mask.shape[-1]), **kwargs) * noindex
        nonzero_idx = torch.where(mask)
        indices[nonzero_idx[0], nonzero_idx[2]] = nonzero_idx[1]
    else:
        raise ValueError("mask must be 2D for single sample or 3D for batch")

    idx_exist = indices >= 0
    minval = torch.min(indices[idx_exist]).item() if idx_exist.any() else 0

    neg_indices = torch.where(indices < 0)
    # ensure indices start from 0
    indices[idx_exist] -= minval
    indices[neg_indices] = noindex

    return indices


def sanitise_mask(
    mask: BoolTensor,
    input_pad_mask: BoolTensor | None = None,
    object_class_preds: Tensor | None = None,
) -> BoolTensor:
    """Sanitise predicted masks by removing padded inputs and null class predictions.

    Parameters
    ----------
    mask : BoolTensor
        The predicted mask
    input_pad_mask : BoolTensor, optional
        The input pad mask, where a value of True respresents a padded input, by default None
    object_class_preds : Tensor, optional
        Object class predictions, by default None
    """
    if input_pad_mask is not None:
        mask.transpose(1, 2)[input_pad_mask] = False
    if object_class_preds is not None:
        pred_null = object_class_preds.argmax(-1) == object_class_preds.shape[-1] - 1
        mask[pred_null] = False
    return mask


def sigmoid_mask(
    mask_logits: Tensor,
    threshold: float = 0.5,
    **kwargs,
) -> BoolTensor:
    """Get a mask by thresholding the mask logits.

    Parameters
    ----------
    mask_logits : Tensor
        The mask logits
    threshold : float, optional
        The threshold, by default 0.5
    **kwargs
        Additional keyword arguments to pass to sanitise_mask
    """
    mask = mask_logits.sigmoid() > threshold
    return sanitise_mask(mask, **kwargs)


def argmax_mask(
    mask_logits: Tensor,
    weighted: bool = False,
    **kwargs,
) -> BoolTensor:
    """Get a mask by taking the argmax of the mask logits.

    Parameters
    ----------
    mask_logits : Tensor
        The mask logits
    weighted : bool, optional
        Weight logits according to object class confidence, as in MaskFormer, by default False
    **kwargs
        Additional keyword arguments to pass to sanitise_mask
    """
    if weighted and kwargs.get("object_class_preds") is None:
        raise ValueError("weighted argmax requires object_class_preds")

    if not weighted:
        idx = mask_logits.argmax(-2)
    else:
        confidence = kwargs["object_class_preds"].max(-1)[0].unsqueeze(-1)
        assert (  # noqa: PT018
            confidence.min() >= 0.0 and confidence.max() <= 1.0
        ), "confidence must be between 0 and 1"
        idx = (mask_logits.softmax(-2) * confidence).argmax(-2)
    mask = mask_from_indices(idx, num_masks=mask_logits.shape[-2])

    return sanitise_mask(mask, **kwargs)


def mask_from_logits(
    logits: Tensor,
    mode: str,
    input_pad_mask: BoolTensor | None = None,
    object_class_preds: Tensor | None = None,
):
    modes = {"sigmoid", "argmax", "weighted_argmax"}
    if mode == "sigmoid":
        return sigmoid_mask(
            logits, input_pad_mask=input_pad_mask, object_class_preds=object_class_preds
        )
    if mode == "argmax":
        return argmax_mask(
            logits, input_pad_mask=input_pad_mask, object_class_preds=object_class_preds
        )
    if mode == "weighted_argmax":
        return argmax_mask(
            logits,
            weighted=True,
            input_pad_mask=input_pad_mask,
            object_class_preds=object_class_preds,
        )

    raise ValueError(f"mode must be one of {modes}")


def mask_effs_purs(m_pred: BoolTensor, m_tgt: BoolTensor) -> tuple[Tensor, Tensor]:
    eff = (m_pred & m_tgt).sum(-1) / m_tgt.sum(-1)
    pur = (m_pred & m_tgt).sum(-1) / m_pred.sum(-1)
    return eff, pur


def mask_eff_pur(m_pred: BoolTensor, m_tgt: BoolTensor, flat: bool = False, reduce: bool = False):
    if flat:
        # per assignment metric (i.e. edgewise)
        eff = (m_pred & m_tgt).sum() / m_tgt.sum()
        pur = (m_pred & m_tgt).sum() / m_pred.sum()
    else:
        # per object metric (nanmean avoids invalid indices)
        eff, pur = mask_effs_purs(m_pred, m_tgt)
        if reduce:
            eff, pur = eff.nanmean(), pur.nanmean()
    return eff, pur


def reco_metrics(
    pred_mask: BoolTensor,
    tgt_mask: BoolTensor,
    pred_valid: Tensor | None = None,
    reduce: bool = False,
    min_recall: float = 1.0,
    min_purity: float = 1.0,
    min_constituents: int = 0,
):
    """Calculate the efficiency and purity of the predicted objects."""
    if pred_valid is None:
        pred_valid = pred_mask.sum(-1) > 0
    else:
        pred_valid = pred_valid.clone()
        pred_valid &= pred_mask.sum(-1) > 0

    eff, pur = mask_effs_purs(pred_mask, tgt_mask)
    pass_cuts = (eff >= min_recall) & (pur >= min_purity)

    if min_constituents > 0:
        pred_valid &= pred_mask.sum(-1) >= min_constituents

    eff = pred_valid & pass_cuts
    fake = pred_valid & ~pass_cuts

    if reduce:
        valid_tgt = tgt_mask.sum(-1) > 0
        eff = eff[valid_tgt].float().mean()
        fake = fake[pred_valid].float().mean()

    return eff, fake
